"""
Template manager for SBYB UI Generator.

This module provides a template manager for UI Generator
to manage and apply templates for different UI frameworks.
"""

from typing import Any, Dict, List, Optional, Union, Tuple
import os
import json
import shutil
import tempfile
import jinja2

from sbyb.core.base import SBYBComponent
from sbyb.core.exceptions import UIGenerationError


class TemplateManager(SBYBComponent):
    """
    Template manager for UI Generator.
    
    This component provides a template manager for UI Generator
    to manage and apply templates for different UI frameworks.
    """
    
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """
        Initialize the template manager.
        
        Args:
            config: Configuration dictionary for the template manager.
        """
        super().__init__(config)
        self.templates = {}
        self.template_dirs = {}
        self._setup_template_dirs()
        self._load_default_templates()
    
    def _setup_template_dirs(self) -> None:
        """
        Set up template directories.
        """
        # Default template directory
        default_template_dir = os.path.join(os.path.dirname(__file__), "templates")
        
        # Create template directory if it doesn't exist
        if not os.path.exists(default_template_dir):
            os.makedirs(default_template_dir)
        
        # Set up template directories for each framework
        for framework in ["streamlit", "dash", "flask", "html", "react", "vue"]:
            framework_dir = os.path.join(default_template_dir, framework)
            if not os.path.exists(framework_dir):
                os.makedirs(framework_dir)
            self.template_dirs[framework] = framework_dir
    
    def _load_default_templates(self) -> None:
        """
        Load default templates.
        """
        # Streamlit templates
        self.register_template(
            "streamlit_basic",
            "Basic Streamlit App",
            "streamlit",
            {
                "description": "A basic Streamlit application template.",
                "files": {
                    "app.py": """
import streamlit as st
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

# Set page config
st.set_page_config(
    page_title="{{ app_title }}",
    page_icon="{{ app_icon }}",
    layout="{{ layout }}",
    initial_sidebar_state="{{ sidebar_state }}"
)

# Title and description
st.title("{{ app_title }}")
st.write("{{ app_description }}")

# Main content
{{ content }}
"""
                }
            }
        )
        
        self.register_template(
            "streamlit_ml_dashboard",
            "ML Dashboard (Streamlit)",
            "streamlit",
            {
                "description": "A Streamlit dashboard for machine learning models.",
                "files": {
                    "app.py": """
import streamlit as st
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pickle
import os
from sklearn.metrics import confusion_matrix, classification_report, roc_curve, auc, precision_recall_curve

# Set page config
st.set_page_config(
    page_title="{{ app_title }}",
    page_icon="{{ app_icon }}",
    layout="wide",
    initial_sidebar_state="expanded"
)

# Title and description
st.title("{{ app_title }}")
st.write("{{ app_description }}")

# Sidebar
with st.sidebar:
    st.header("Navigation")
    page = st.radio("Go to", ["Overview", "Model Performance", "Predictions", "About"])

# Load data and model
@st.cache_data
def load_data():
    return pd.read_csv("{{ data_path }}")

@st.cache_resource
def load_model():
    with open("{{ model_path }}", "rb") as f:
        return pickle.load(f)

try:
    data = load_data()
    model = load_model()
except Exception as e:
    st.error(f"Error loading data or model: {str(e)}")
    st.stop()

# Pages
if page == "Overview":
    st.header("Dataset Overview")
    
    # Display dataset info
    st.subheader("Dataset Information")
    st.write(f"Number of rows: {data.shape[0]}")
    st.write(f"Number of columns: {data.shape[1]}")
    
    # Display first few rows
    st.subheader("Sample Data")
    st.dataframe(data.head())
    
    # Display summary statistics
    st.subheader("Summary Statistics")
    st.dataframe(data.describe())
    
    # Display visualizations
    st.subheader("Visualizations")
    
    # Select columns for visualization
    col1, col2 = st.columns(2)
    with col1:
        numeric_cols = data.select_dtypes(include=["number"]).columns.tolist()
        x_col = st.selectbox("Select X-axis", numeric_cols)
    with col2:
        y_col = st.selectbox("Select Y-axis", numeric_cols)
    
    # Create visualization
    fig, ax = plt.subplots(figsize=(10, 6))
    sns.scatterplot(data=data, x=x_col, y=y_col, ax=ax)
    ax.set_title(f"{y_col} vs {x_col}")
    st.pyplot(fig)

elif page == "Model Performance":
    st.header("Model Performance")
    
    # Get target column
    target_col = "{{ target_column }}"
    
    # Get feature columns
    feature_cols = [col for col in data.columns if col != target_col]
    
    # Split data
    X = data[feature_cols]
    y = data[target_col]
    
    # Make predictions
    y_pred = model.predict(X)
    
    # Classification metrics
    if hasattr(model, "predict_proba"):
        st.subheader("Classification Metrics")
        
        # Confusion Matrix
        st.write("Confusion Matrix")
        cm = confusion_matrix(y, y_pred)
        fig, ax = plt.subplots(figsize=(8, 6))
        sns.heatmap(cm, annot=True, fmt="d", cmap="Blues", ax=ax)
        ax.set_xlabel("Predicted")
        ax.set_ylabel("True")
        st.pyplot(fig)
        
        # Classification Report
        st.write("Classification Report")
        report = classification_report(y, y_pred, output_dict=True)
        report_df = pd.DataFrame(report).transpose()
        st.dataframe(report_df)
        
        # ROC Curve
        st.write("ROC Curve")
        y_proba = model.predict_proba(X)
        
        # For binary classification
        if y_proba.shape[1] == 2:
            fpr, tpr, _ = roc_curve(y, y_proba[:, 1])
            roc_auc = auc(fpr, tpr)
            
            fig, ax = plt.subplots(figsize=(8, 6))
            ax.plot(fpr, tpr, label=f"ROC curve (area = {roc_auc:.2f})")
            ax.plot([0, 1], [0, 1], "k--")
            ax.set_xlim([0.0, 1.0])
            ax.set_ylim([0.0, 1.05])
            ax.set_xlabel("False Positive Rate")
            ax.set_ylabel("True Positive Rate")
            ax.set_title("Receiver Operating Characteristic")
            ax.legend(loc="lower right")
            st.pyplot(fig)
    
    # Regression metrics
    else:
        st.subheader("Regression Metrics")
        
        # Residuals
        residuals = y - y_pred
        
        # Residual plot
        st.write("Residual Plot")
        fig, ax = plt.subplots(figsize=(8, 6))
        ax.scatter(y_pred, residuals)
        ax.axhline(y=0, color="r", linestyle="-")
        ax.set_xlabel("Predicted Values")
        ax.set_ylabel("Residuals")
        ax.set_title("Residuals vs Predicted Values")
        st.pyplot(fig)
        
        # Residual histogram
        st.write("Residual Histogram")
        fig, ax = plt.subplots(figsize=(8, 6))
        ax.hist(residuals, bins=20, edgecolor="black")
        ax.axvline(x=0, color="r", linestyle="-")
        ax.set_xlabel("Residuals")
        ax.set_ylabel("Frequency")
        ax.set_title("Histogram of Residuals")
        st.pyplot(fig)
        
        # Actual vs Predicted
        st.write("Actual vs Predicted")
        fig, ax = plt.subplots(figsize=(8, 6))
        ax.scatter(y, y_pred)
        ax.plot([y.min(), y.max()], [y.min(), y.max()], "k--")
        ax.set_xlabel("Actual Values")
        ax.set_ylabel("Predicted Values")
        ax.set_title("Actual vs Predicted Values")
        st.pyplot(fig)

elif page == "Predictions":
    st.header("Make Predictions")
    
    # Get feature columns
    feature_cols = [col for col in data.columns if col != "{{ target_column }}"]
    
    # Create input form
    st.subheader("Input Features")
    
    # Create columns for input fields
    cols = st.columns(2)
    input_data = {}
    
    for i, col in enumerate(feature_cols):
        with cols[i % 2]:
            # Determine input type based on data type
            if data[col].dtype == "float64":
                input_data[col] = st.number_input(f"{col}", value=float(data[col].mean()))
            elif data[col].dtype == "int64":
                input_data[col] = st.number_input(f"{col}", value=int(data[col].mean()), step=1)
            elif data[col].nunique() < 10:  # Categorical with few values
                input_data[col] = st.selectbox(f"{col}", options=data[col].unique())
            else:
                input_data[col] = st.text_input(f"{col}", value=str(data[col].iloc[0]))
    
    # Make prediction
    if st.button("Predict"):
        try:
            # Create input DataFrame
            input_df = pd.DataFrame([input_data])
            
            # Make prediction
            prediction = model.predict(input_df)
            
            # Display prediction
            st.success(f"Prediction: {prediction[0]}")
            
            # Display probability if available
            if hasattr(model, "predict_proba"):
                proba = model.predict_proba(input_df)
                st.write("Prediction Probabilities:")
                
                # Get class names if available
                if hasattr(model, "classes_"):
                    class_names = model.classes_
                    proba_df = pd.DataFrame([proba[0]], columns=class_names)
                else:
                    proba_df = pd.DataFrame([proba[0]], columns=[f"Class {i}" for i in range(proba.shape[1])])
                
                st.dataframe(proba_df)
        except Exception as e:
            st.error(f"Error making prediction: {str(e)}")

else:  # About page
    st.header("About")
    
    st.subheader("Application Information")
    st.write("{{ app_description }}")
    
    st.subheader("Model Information")
    st.write(f"Model Type: {type(model).__name__}")
    
    # Display model parameters if available
    if hasattr(model, "get_params"):
        st.write("Model Parameters:")
        params = model.get_params()
        params_df = pd.DataFrame({"Parameter": list(params.keys()), "Value": list(params.values())})
        st.dataframe(params_df)
    
    st.subheader("Dataset Information")
    st.write(f"Dataset Shape: {data.shape}")
    st.write(f"Features: {', '.join([col for col in data.columns if col != '{{ target_column }}'])}")
    st.write(f"Target: {{ target_column }}")
"""
                }
            }
        )
        
        # Dash templates
        self.register_template(
            "dash_basic",
            "Basic Dash App",
            "dash",
            {
                "description": "A basic Dash application template.",
                "files": {
                    "app.py": """
import dash
from dash import dcc, html, callback, Input, Output
import pandas as pd
import plotly.express as px

# Initialize the Dash app
app = dash.Dash(__name__, title="{{ app_title }}")

# Define the layout
app.layout = html.Div([
    html.H1("{{ app_title }}"),
    html.P("{{ app_description }}"),
    
    # Main content
    {{ content }}
])

# Run the app
if __name__ == "__main__":
    app.run_server(debug=True, host="0.0.0.0")
"""
                }
            }
        )
        
        self.register_template(
            "dash_ml_dashboard",
            "ML Dashboard (Dash)",
            "dash",
            {
                "description": "A Dash dashboard for machine learning models.",
                "files": {
                    "app.py": """
import dash
from dash import dcc, html, callback, Input, Output, State, dash_table
import pandas as pd
import numpy as np
import plotly.express as px
import plotly.graph_objects as go
import pickle
import os
from sklearn.metrics import confusion_matrix, classification_report, roc_curve, auc, precision_recall_curve

# Initialize the Dash app
app = dash.Dash(__name__, title="{{ app_title }}")

# Load data and model
def load_data():
    return pd.read_csv("{{ data_path }}")

def load_model():
    with open("{{ model_path }}", "rb") as f:
        return pickle.load(f)

try:
    data = load_data()
    model = load_model()
except Exception as e:
    print(f"Error loading data or model: {str(e)}")

# Get target column
target_col = "{{ target_column }}"

# Get feature columns
feature_cols = [col for col in data.columns if col != target_col]

# Define the layout
app.layout = html.Div([
    # Header
    html.H1("{{ app_title }}"),
    html.P("{{ app_description }}"),
    
    # Tabs
    dcc.Tabs([
        # Overview Tab
        dcc.Tab(label="Overview", children=[
            html.Div([
                html.H2("Dataset Overview"),
                
                # Dataset info
                html.H3("Dataset Information"),
                html.P(f"Number of rows: {data.shape[0]}"),
                html.P(f"Number of columns: {data.shape[1]}"),
                
                # Sample data
                html.H3("Sample Data"),
                dash_table.DataTable(
                    data=data.head().to_dict("records"),
                    columns=[{"name": i, "id": i} for i in data.columns],
                    page_size=5,
                    style_table={"overflowX": "auto"}
                ),
                
                # Summary statistics
                html.H3("Summary Statistics"),
                dash_table.DataTable(
                    data=data.describe().reset_index().to_dict("records"),
                    columns=[{"name": i, "id": i} for i in data.describe().reset_index().columns],
                    page_size=10,
                    style_table={"overflowX": "auto"}
                ),
                
                # Visualizations
                html.H3("Visualizations"),
                
                # Controls for visualization
                html.Div([
                    html.Div([
                        html.Label("Select X-axis"),
                        dcc.Dropdown(
                            id="x-column",
                            options=[{"label": col, "value": col} for col in data.select_dtypes(include=["number"]).columns],
                            value=data.select_dtypes(include=["number"]).columns[0]
                        )
                    ], style={"width": "48%", "display": "inline-block"}),
                    
                    html.Div([
                        html.Label("Select Y-axis"),
                        dcc.Dropdown(
                            id="y-column",
                            options=[{"label": col, "value": col} for col in data.select_dtypes(include=["number"]).columns],
                            value=data.select_dtypes(include=["number"]).columns[1] if len(data.select_dtypes(include=["number"]).columns) > 1 else data.select_dtypes(include=["number"]).columns[0]
                        )
                    ], style={"width": "48%", "float": "right", "display": "inline-block"})
                ]),
                
                # Scatter plot
                dcc.Graph(id="scatter-plot")
            ])
        ]),
        
        # Model Performance Tab
        dcc.Tab(label="Model Performance", children=[
            html.Div([
                html.H2("Model Performance"),
                
                # Classification metrics
                html.Div(id="classification-metrics", children=[
                    html.H3("Classification Metrics"),
                    
                    # Confusion Matrix
                    html.H4("Confusion Matrix"),
                    dcc.Graph(id="confusion-matrix"),
                    
                    # Classification Report
                    html.H4("Classification Report"),
                    html.Div(id="classification-report"),
                    
                    # ROC Curve
                    html.H4("ROC Curve"),
                    dcc.Graph(id="roc-curve")
                ]) if hasattr(model, "predict_proba") else [],
                
                # Regression metrics
                html.Div(id="regression-metrics", children=[
                    html.H3("Regression Metrics"),
                    
                    # Residual Plot
                    html.H4("Residual Plot"),
                    dcc.Graph(id="residual-plot"),
                    
                    # Residual Histogram
                    html.H4("Residual Histogram"),
                    dcc.Graph(id="residual-histogram"),
                    
                    # Actual vs Predicted
                    html.H4("Actual vs Predicted"),
                    dcc.Graph(id="actual-vs-predicted")
                ]) if not hasattr(model, "predict_proba") else []
            ])
        ]),
        
        # Predictions Tab
        dcc.Tab(label="Predictions", children=[
            html.Div([
                html.H2("Make Predictions"),
                
                # Input form
                html.H3("Input Features"),
                
                # Create input fields
                html.Div([
                    html.Div([
                        html.Div([
                            html.Label(col),
                            dcc.Input(
                                id=f"input-{col}",
                                type="number",
                                value=float(data[col].mean()) if data[col].dtype in ["float64", "int64"] else 0,
                                step=1 if data[col].dtype == "int64" else 0.1
                            )
                        ]) for col in feature_cols[i::2]
                    ], style={"width": "48%", "display": "inline-block"}) for i in range(2)
                ]),
                
                # Predict button
                html.Button("Predict", id="predict-button", n_clicks=0),
                
                # Prediction output
                html.Div(id="prediction-output")
            ])
        ]),
        
        # About Tab
        dcc.Tab(label="About", children=[
            html.Div([
                html.H2("About"),
                
                # Application Information
                html.H3("Application Information"),
                html.P("{{ app_description }}"),
                
                # Model Information
                html.H3("Model Information"),
                html.P(f"Model Type: {type(model).__name__}"),
                
                # Model Parameters
                html.H4("Model Parameters"),
                html.Div(id="model-parameters"),
                
                # Dataset Information
                html.H3("Dataset Information"),
                html.P(f"Dataset Shape: {data.shape}"),
                html.P(f"Features: {', '.join([col for col in data.columns if col != target_col])}"),
                html.P(f"Target: {target_col}")
            ])
        ])
    ])
])

# Callbacks
@callback(
    Output("scatter-plot", "figure"),
    [Input("x-column", "value"), Input("y-column", "value")]
)
def update_scatter_plot(x_col, y_col):
    fig = px.scatter(data, x=x_col, y=y_col, title=f"{y_col} vs {x_col}")
    return fig

# Classification metrics callbacks
if hasattr(model, "predict_proba"):
    # Confusion Matrix
    @callback(
        Output("confusion-matrix", "figure"),
        [Input("classification-metrics", "children")]
    )
    def update_confusion_matrix(_):
        # Make predictions
        y_pred = model.predict(data[feature_cols])
        
        # Compute confusion matrix
        cm = confusion_matrix(data[target_col], y_pred)
        
        # Create heatmap
        fig = px.imshow(
            cm,
            labels=dict(x="Predicted", y="True", color="Count"),
            x=model.classes_ if hasattr(model, "classes_") else [f"Class {i}" for i in range(cm.shape[1])],
            y=model.classes_ if hasattr(model, "classes_") else [f"Class {i}" for i in range(cm.shape[0])],
            text_auto=True,
            color_continuous_scale="Blues"
        )
        
        return fig
    
    # Classification Report
    @callback(
        Output("classification-report", "children"),
        [Input("classification-metrics", "children")]
    )
    def update_classification_report(_):
        # Make predictions
        y_pred = model.predict(data[feature_cols])
        
        # Compute classification report
        report = classification_report(data[target_col], y_pred, output_dict=True)
        report_df = pd.DataFrame(report).transpose()
        
        # Create table
        return dash_table.DataTable(
            data=report_df.reset_index().to_dict("records"),
            columns=[{"name": i, "id": i} for i in report_df.reset_index().columns],
            page_size=10,
            style_table={"overflowX": "auto"}
        )
    
    # ROC Curve
    @callback(
        Output("roc-curve", "figure"),
        [Input("classification-metrics", "children")]
    )
    def update_roc_curve(_):
        # Make predictions
        y_proba = model.predict_proba(data[feature_cols])
        
        # For binary classification
        if y_proba.shape[1] == 2:
            fpr, tpr, _ = roc_curve(data[target_col], y_proba[:, 1])
            roc_auc = auc(fpr, tpr)
            
            fig = go.Figure()
            fig.add_trace(go.Scatter(x=fpr, y=tpr, mode="lines", name=f"ROC curve (area = {roc_auc:.2f})"))
            fig.add_trace(go.Scatter(x=[0, 1], y=[0, 1], mode="lines", name="Random", line=dict(dash="dash")))
            
            fig.update_layout(
                title="Receiver Operating Characteristic",
                xaxis_title="False Positive Rate",
                yaxis_title="True Positive Rate",
                legend=dict(x=0.7, y=0.1),
                xaxis=dict(range=[0, 1]),
                yaxis=dict(range=[0, 1.05])
            )
            
            return fig
        else:
            # For multiclass, return empty figure
            return go.Figure()

# Regression metrics callbacks
else:
    # Residual Plot
    @callback(
        Output("residual-plot", "figure"),
        [Input("regression-metrics", "children")]
    )
    def update_residual_plot(_):
        # Make predictions
        y_pred = model.predict(data[feature_cols])
        
        # Compute residuals
        residuals = data[target_col] - y_pred
        
        # Create scatter plot
        fig = px.scatter(
            x=y_pred,
            y=residuals,
            labels={"x": "Predicted Values", "y": "Residuals"},
            title="Residuals vs Predicted Values"
        )
        
        # Add horizontal line at y=0
        fig.add_shape(
            type="line",
            x0=min(y_pred),
            y0=0,
            x1=max(y_pred),
            y1=0,
            line=dict(color="red", width=2, dash="dash")
        )
        
        return fig
    
    # Residual Histogram
    @callback(
        Output("residual-histogram", "figure"),
        [Input("regression-metrics", "children")]
    )
    def update_residual_histogram(_):
        # Make predictions
        y_pred = model.predict(data[feature_cols])
        
        # Compute residuals
        residuals = data[target_col] - y_pred
        
        # Create histogram
        fig = px.histogram(
            residuals,
            labels={"value": "Residuals", "count": "Frequency"},
            title="Histogram of Residuals"
        )
        
        # Add vertical line at x=0
        fig.add_shape(
            type="line",
            x0=0,
            y0=0,
            x1=0,
            y1=max(fig.data[0].y),
            line=dict(color="red", width=2, dash="dash")
        )
        
        return fig
    
    # Actual vs Predicted
    @callback(
        Output("actual-vs-predicted", "figure"),
        [Input("regression-metrics", "children")]
    )
    def update_actual_vs_predicted(_):
        # Make predictions
        y_pred = model.predict(data[feature_cols])
        
        # Create scatter plot
        fig = px.scatter(
            x=data[target_col],
            y=y_pred,
            labels={"x": "Actual Values", "y": "Predicted Values"},
            title="Actual vs Predicted Values"
        )
        
        # Add diagonal line
        min_val = min(min(data[target_col]), min(y_pred))
        max_val = max(max(data[target_col]), max(y_pred))
        
        fig.add_shape(
            type="line",
            x0=min_val,
            y0=min_val,
            x1=max_val,
            y1=max_val,
            line=dict(color="red", width=2, dash="dash")
        )
        
        return fig

# Prediction callback
@callback(
    Output("prediction-output", "children"),
    [Input("predict-button", "n_clicks")],
    [State(f"input-{col}", "value") for col in feature_cols]
)
def make_prediction(n_clicks, *feature_values):
    if n_clicks == 0:
        return html.Div()
    
    try:
        # Create input dictionary
        input_data = {col: val for col, val in zip(feature_cols, feature_values)}
        
        # Create input DataFrame
        input_df = pd.DataFrame([input_data])
        
        # Make prediction
        prediction = model.predict(input_df)
        
        # Create output
        output = [html.Div(html.H3(f"Prediction: {prediction[0]}"), style={"color": "green"})]
        
        # Add probability if available
        if hasattr(model, "predict_proba"):
            proba = model.predict_proba(input_df)
            
            # Get class names if available
            if hasattr(model, "classes_"):
                class_names = model.classes_
                proba_df = pd.DataFrame([proba[0]], columns=class_names)
            else:
                proba_df = pd.DataFrame([proba[0]], columns=[f"Class {i}" for i in range(proba.shape[1])])
            
            output.append(html.H4("Prediction Probabilities:"))
            output.append(dash_table.DataTable(
                data=proba_df.to_dict("records"),
                columns=[{"name": str(i), "id": str(i)} for i in proba_df.columns],
                style_table={"overflowX": "auto"}
            ))
        
        return html.Div(output)
    except Exception as e:
        return html.Div(html.H3(f"Error: {str(e)}"), style={"color": "red"})

# Model parameters callback
@callback(
    Output("model-parameters", "children"),
    [Input("model-parameters", "children")]
)
def update_model_parameters(_):
    if hasattr(model, "get_params"):
        params = model.get_params()
        params_df = pd.DataFrame({"Parameter": list(params.keys()), "Value": list(params.values())})
        
        return dash_table.DataTable(
            data=params_df.to_dict("records"),
            columns=[{"name": i, "id": i} for i in params_df.columns],
            page_size=10,
            style_table={"overflowX": "auto"}
        )
    else:
        return html.P("Model parameters not available.")

# Run the app
if __name__ == "__main__":
    app.run_server(debug=True, host="0.0.0.0")
"""
                }
            }
        )
        
        # Flask templates
        self.register_template(
            "flask_basic",
            "Basic Flask App",
            "flask",
            {
                "description": "A basic Flask application template.",
                "files": {
                    "app.py": """
from flask import Flask, render_template, request, jsonify

app = Flask(__name__)

@app.route("/")
def index():
    return render_template("index.html", title="{{ app_title }}")

{{ content }}

if __name__ == "__main__":
    app.run(debug=True, host="0.0.0.0")
""",
                    "templates/index.html": """
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>{{ title }}</title>
    <link rel="stylesheet" href="{{ url_for('static', filename='css/style.css') }}">
</head>
<body>
    <header>
        <h1>{{ title }}</h1>
        <p>{{ app_description }}</p>
    </header>
    
    <main>
        <!-- Main content will go here -->
    </main>
    
    <footer>
        <p>&copy; {{ current_year }} {{ app_title }}</p>
    </footer>
    
    <script src="{{ url_for('static', filename='js/script.js') }}"></script>
</body>
</html>
""",
                    "static/css/style.css": """
/* Basic styles */
body {
    font-family: Arial, sans-serif;
    line-height: 1.6;
    margin: 0;
    padding: 0;
    color: #333;
}

header {
    background-color: #f4f4f4;
    padding: 1rem;
    text-align: center;
}

main {
    padding: 1rem;
    max-width: 1200px;
    margin: 0 auto;
}

footer {
    background-color: #f4f4f4;
    padding: 1rem;
    text-align: center;
    margin-top: 2rem;
}
""",
                    "static/js/script.js": """
// JavaScript for the application
document.addEventListener('DOMContentLoaded', function() {
    console.log('Application loaded');
});
"""
                }
            }
        )
        
        self.register_template(
            "flask_ml_api",
            "ML API (Flask)",
            "flask",
            {
                "description": "A Flask API for machine learning models.",
                "files": {
                    "app.py": """
from flask import Flask, request, jsonify, render_template
import pandas as pd
import numpy as np
import pickle
import os
import json

app = Flask(__name__)

# Load model
def load_model():
    with open("{{ model_path }}", "rb") as f:
        return pickle.load(f)

try:
    model = load_model()
    print(f"Model loaded: {type(model).__name__}")
except Exception as e:
    print(f"Error loading model: {str(e)}")
    model = None

# Get feature names
try:
    if hasattr(model, "feature_names_in_"):
        feature_names = model.feature_names_in_.tolist()
    else:
        # Try to load feature names from a file
        try:
            with open("{{ feature_names_path }}", "r") as f:
                feature_names = json.load(f)
        except:
            feature_names = []
except:
    feature_names = []

@app.route("/")
def index():
    return render_template("index.html", 
                          title="{{ app_title }}", 
                          description="{{ app_description }}",
                          feature_names=feature_names)

@app.route("/api/predict", methods=["POST"])
def predict():
    if model is None:
        return jsonify({"error": "Model not loaded"}), 500
    
    try:
        # Get input data
        data = request.json
        
        # Validate input
        if not data or not isinstance(data, dict):
            return jsonify({"error": "Invalid input format. Expected JSON object."}), 400
        
        # Create input DataFrame
        input_df = pd.DataFrame([data])
        
        # Check if all required features are present
        if feature_names and not all(feature in input_df.columns for feature in feature_names):
            missing_features = [feature for feature in feature_names if feature not in input_df.columns]
            return jsonify({"error": f"Missing features: {', '.join(missing_features)}"}), 400
        
        # Make prediction
        prediction = model.predict(input_df)
        
        # Prepare response
        response = {
            "prediction": prediction[0].tolist() if isinstance(prediction[0], (np.ndarray, list)) else float(prediction[0])
        }
        
        # Add probability if available
        if hasattr(model, "predict_proba"):
            proba = model.predict_proba(input_df)
            
            # Get class names if available
            if hasattr(model, "classes_"):
                class_names = model.classes_.tolist()
                response["probabilities"] = {str(class_name): float(prob) for class_name, prob in zip(class_names, proba[0])}
            else:
                response["probabilities"] = {f"class_{i}": float(prob) for i, prob in enumerate(proba[0])}
        
        return jsonify(response)
    except Exception as e:
        return jsonify({"error": str(e)}), 500

@app.route("/api/model-info")
def model_info():
    if model is None:
        return jsonify({"error": "Model not loaded"}), 500
    
    try:
        # Get model information
        info = {
            "model_type": type(model).__name__,
            "feature_names": feature_names
        }
        
        # Add model parameters if available
        if hasattr(model, "get_params"):
            params = model.get_params()
            info["parameters"] = {k: str(v) for k, v in params.items()}
        
        return jsonify(info)
    except Exception as e:
        return jsonify({"error": str(e)}), 500

if __name__ == "__main__":
    app.run(debug=True, host="0.0.0.0")
""",
                    "templates/index.html": """
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>{{ title }}</title>
    <link rel="stylesheet" href="{{ url_for('static', filename='css/style.css') }}">
</head>
<body>
    <header>
        <h1>{{ title }}</h1>
        <p>{{ description }}</p>
    </header>
    
    <main>
        <section class="api-docs">
            <h2>API Documentation</h2>
            
            <div class="endpoint">
                <h3>Prediction Endpoint</h3>
                <p><strong>URL:</strong> <code>/api/predict</code></p>
                <p><strong>Method:</strong> POST</p>
                <p><strong>Content-Type:</strong> application/json</p>
                
                <h4>Request Body:</h4>
                <pre><code>{
    {% for feature in feature_names %}
    "{{ feature }}": value{% if not loop.last %},{% endif %}
    {% endfor %}
}</code></pre>
                
                <h4>Response:</h4>
                <pre><code>{
    "prediction": value
    {% if has_probabilities %}
    "probabilities": {
        "class_1": probability_1,
        "class_2": probability_2,
        ...
    }
    {% endif %}
}</code></pre>
            </div>
            
            <div class="endpoint">
                <h3>Model Information Endpoint</h3>
                <p><strong>URL:</strong> <code>/api/model-info</code></p>
                <p><strong>Method:</strong> GET</p>
                
                <h4>Response:</h4>
                <pre><code>{
    "model_type": "ModelType",
    "feature_names": ["feature1", "feature2", ...],
    "parameters": {
        "param1": "value1",
        "param2": "value2",
        ...
    }
}</code></pre>
            </div>
        </section>
        
        <section class="test-api">
            <h2>Test API</h2>
            
            <div class="input-form">
                <h3>Input Features</h3>
                
                <form id="prediction-form">
                    {% for feature in feature_names %}
                    <div class="form-group">
                        <label for="{{ feature }}">{{ feature }}</label>
                        <input type="number" id="{{ feature }}" name="{{ feature }}" step="0.01" required>
                    </div>
                    {% endfor %}
                    
                    <button type="submit">Predict</button>
                </form>
            </div>
            
            <div class="output-result">
                <h3>Prediction Result</h3>
                <div id="prediction-result"></div>
            </div>
        </section>
    </main>
    
    <footer>
        <p>&copy; 2023 {{ title }}</p>
    </footer>
    
    <script src="{{ url_for('static', filename='js/script.js') }}"></script>
</body>
</html>
""",
                    "static/css/style.css": """
/* Basic styles */
body {
    font-family: Arial, sans-serif;
    line-height: 1.6;
    margin: 0;
    padding: 0;
    color: #333;
}

header {
    background-color: #f4f4f4;
    padding: 1rem;
    text-align: center;
}

main {
    padding: 1rem;
    max-width: 1200px;
    margin: 0 auto;
}

footer {
    background-color: #f4f4f4;
    padding: 1rem;
    text-align: center;
    margin-top: 2rem;
}

/* API Documentation */
.api-docs {
    margin-bottom: 2rem;
}

.endpoint {
    background-color: #f9f9f9;
    padding: 1rem;
    margin-bottom: 1rem;
    border-radius: 5px;
}

pre {
    background-color: #f1f1f1;
    padding: 1rem;
    border-radius: 5px;
    overflow-x: auto;
}

code {
    font-family: monospace;
}

/* Test API */
.test-api {
    display: flex;
    flex-wrap: wrap;
    gap: 2rem;
}

.input-form, .output-result {
    flex: 1;
    min-width: 300px;
}

.form-group {
    margin-bottom: 1rem;
}

label {
    display: block;
    margin-bottom: 0.5rem;
}

input {
    width: 100%;
    padding: 0.5rem;
    border: 1px solid #ddd;
    border-radius: 3px;
}

button {
    background-color: #4CAF50;
    color: white;
    padding: 0.5rem 1rem;
    border: none;
    border-radius: 3px;
    cursor: pointer;
}

button:hover {
    background-color: #45a049;
}

#prediction-result {
    background-color: #f9f9f9;
    padding: 1rem;
    border-radius: 5px;
    min-height: 100px;
}
""",
                    "static/js/script.js": """
// JavaScript for the application
document.addEventListener('DOMContentLoaded', function() {
    console.log('Application loaded');
    
    // Get prediction form
    const predictionForm = document.getElementById('prediction-form');
    
    // Add submit event listener
    if (predictionForm) {
        predictionForm.addEventListener('submit', function(e) {
            e.preventDefault();
            
            // Get form data
            const formData = new FormData(predictionForm);
            const data = {};
            
            // Convert form data to JSON
            for (const [key, value] of formData.entries()) {
                data[key] = parseFloat(value);
            }
            
            // Make prediction request
            fetch('/api/predict', {
                method: 'POST',
                headers: {
                    'Content-Type': 'application/json'
                },
                body: JSON.stringify(data)
            })
            .then(response => response.json())
            .then(result => {
                // Display prediction result
                const predictionResult = document.getElementById('prediction-result');
                
                if (result.error) {
                    predictionResult.innerHTML = `<div class="error">${result.error}</div>`;
                } else {
                    let html = `<div class="prediction">Prediction: <strong>${result.prediction}</strong></div>`;
                    
                    // Display probabilities if available
                    if (result.probabilities) {
                        html += '<div class="probabilities"><h4>Probabilities:</h4><ul>';
                        
                        for (const [className, probability] of Object.entries(result.probabilities)) {
                            html += `<li>${className}: ${(probability * 100).toFixed(2)}%</li>`;
                        }
                        
                        html += '</ul></div>';
                    }
                    
                    predictionResult.innerHTML = html;
                }
            })
            .catch(error => {
                console.error('Error:', error);
                const predictionResult = document.getElementById('prediction-result');
                predictionResult.innerHTML = `<div class="error">Error: ${error.message}</div>`;
            });
        });
    }
});
"""
                }
            }
        )
        
        # HTML templates
        self.register_template(
            "html_basic",
            "Basic HTML Page",
            "html",
            {
                "description": "A basic HTML page template.",
                "files": {
                    "index.html": """
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>{{ app_title }}</title>
    <link rel="stylesheet" href="css/style.css">
</head>
<body>
    <header>
        <h1>{{ app_title }}</h1>
        <p>{{ app_description }}</p>
    </header>
    
    <main>
        <!-- Main content will go here -->
        {{ content }}
    </main>
    
    <footer>
        <p>&copy; 2023 {{ app_title }}</p>
    </footer>
    
    <script src="js/script.js"></script>
</body>
</html>
""",
                    "css/style.css": """
/* Basic styles */
body {
    font-family: Arial, sans-serif;
    line-height: 1.6;
    margin: 0;
    padding: 0;
    color: #333;
}

header {
    background-color: #f4f4f4;
    padding: 1rem;
    text-align: center;
}

main {
    padding: 1rem;
    max-width: 1200px;
    margin: 0 auto;
}

footer {
    background-color: #f4f4f4;
    padding: 1rem;
    text-align: center;
    margin-top: 2rem;
}
""",
                    "js/script.js": """
// JavaScript for the application
document.addEventListener('DOMContentLoaded', function() {
    console.log('Application loaded');
});
"""
                }
            }
        )
        
        # React templates
        self.register_template(
            "react_basic",
            "Basic React App",
            "react",
            {
                "description": "A basic React application template.",
                "files": {
                    "package.json": """
{
  "name": "{{ app_name }}",
  "version": "0.1.0",
  "private": true,
  "dependencies": {
    "@testing-library/jest-dom": "^5.16.5",
    "@testing-library/react": "^13.4.0",
    "@testing-library/user-event": "^13.5.0",
    "react": "^18.2.0",
    "react-dom": "^18.2.0",
    "react-scripts": "5.0.1",
    "web-vitals": "^2.1.4"
  },
  "scripts": {
    "start": "react-scripts start",
    "build": "react-scripts build",
    "test": "react-scripts test",
    "eject": "react-scripts eject"
  },
  "eslintConfig": {
    "extends": [
      "react-app",
      "react-app/jest"
    ]
  },
  "browserslist": {
    "production": [
      ">0.2%",
      "not dead",
      "not op_mini all"
    ],
    "development": [
      "last 1 chrome version",
      "last 1 firefox version",
      "last 1 safari version"
    ]
  }
}
""",
                    "public/index.html": """
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8" />
    <link rel="icon" href="%PUBLIC_URL%/favicon.ico" />
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <meta name="theme-color" content="#000000" />
    <meta
      name="description"
      content="{{ app_description }}"
    />
    <link rel="apple-touch-icon" href="%PUBLIC_URL%/logo192.png" />
    <link rel="manifest" href="%PUBLIC_URL%/manifest.json" />
    <title>{{ app_title }}</title>
  </head>
  <body>
    <noscript>You need to enable JavaScript to run this app.</noscript>
    <div id="root"></div>
  </body>
</html>
""",
                    "src/index.js": """
import React from 'react';
import ReactDOM from 'react-dom/client';
import './index.css';
import App from './App';
import reportWebVitals from './reportWebVitals';

const root = ReactDOM.createRoot(document.getElementById('root'));
root.render(
  <React.StrictMode>
    <App />
  </React.StrictMode>
);

reportWebVitals();
""",
                    "src/App.js": """
import React from 'react';
import './App.css';

function App() {
  return (
    <div className="App">
      <header className="App-header">
        <h1>{{ app_title }}</h1>
        <p>{{ app_description }}</p>
      </header>
      
      <main className="App-main">
        {/* Main content will go here */}
        {{ content }}
      </main>
      
      <footer className="App-footer">
        <p>&copy; 2023 {{ app_title }}</p>
      </footer>
    </div>
  );
}

export default App;
""",
                    "src/App.css": """
.App {
  text-align: center;
}

.App-header {
  background-color: #282c34;
  min-height: 20vh;
  display: flex;
  flex-direction: column;
  align-items: center;
  justify-content: center;
  font-size: calc(10px + 2vmin);
  color: white;
  padding: 1rem;
}

.App-main {
  padding: 2rem;
  max-width: 1200px;
  margin: 0 auto;
}

.App-footer {
  background-color: #f4f4f4;
  padding: 1rem;
  text-align: center;
  margin-top: 2rem;
}
""",
                    "src/index.css": """
body {
  margin: 0;
  font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', 'Roboto', 'Oxygen',
    'Ubuntu', 'Cantarell', 'Fira Sans', 'Droid Sans', 'Helvetica Neue',
    sans-serif;
  -webkit-font-smoothing: antialiased;
  -moz-osx-font-smoothing: grayscale;
}

code {
  font-family: source-code-pro, Menlo, Monaco, Consolas, 'Courier New',
    monospace;
}
""",
                    "src/reportWebVitals.js": """
const reportWebVitals = (onPerfEntry) => {
  if (onPerfEntry && onPerfEntry instanceof Function) {
    import('web-vitals').then(({ getCLS, getFID, getFCP, getLCP, getTTFB }) => {
      getCLS(onPerfEntry);
      getFID(onPerfEntry);
      getFCP(onPerfEntry);
      getLCP(onPerfEntry);
      getTTFB(onPerfEntry);
    });
  }
};

export default reportWebVitals;
"""
                }
            }
        )
        
        # Vue templates
        self.register_template(
            "vue_basic",
            "Basic Vue App",
            "vue",
            {
                "description": "A basic Vue application template.",
                "files": {
                    "package.json": """
{
  "name": "{{ app_name }}",
  "version": "0.1.0",
  "private": true,
  "scripts": {
    "serve": "vue-cli-service serve",
    "build": "vue-cli-service build",
    "lint": "vue-cli-service lint"
  },
  "dependencies": {
    "core-js": "^3.8.3",
    "vue": "^3.2.13"
  },
  "devDependencies": {
    "@babel/core": "^7.12.16",
    "@babel/eslint-parser": "^7.12.16",
    "@vue/cli-plugin-babel": "~5.0.0",
    "@vue/cli-plugin-eslint": "~5.0.0",
    "@vue/cli-service": "~5.0.0",
    "eslint": "^7.32.0",
    "eslint-plugin-vue": "^8.0.3"
  },
  "eslintConfig": {
    "root": true,
    "env": {
      "node": true
    },
    "extends": [
      "plugin:vue/vue3-essential",
      "eslint:recommended"
    ],
    "parserOptions": {
      "parser": "@babel/eslint-parser"
    },
    "rules": {}
  },
  "browserslist": [
    "> 1%",
    "last 2 versions",
    "not dead",
    "not ie 11"
  ]
}
""",
                    "public/index.html": """
<!DOCTYPE html>
<html lang="">
  <head>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width,initial-scale=1.0">
    <link rel="icon" href="<%= BASE_URL %>favicon.ico">
    <title>{{ app_title }}</title>
  </head>
  <body>
    <noscript>
      <strong>We're sorry but {{ app_title }} doesn't work properly without JavaScript enabled. Please enable it to continue.</strong>
    </noscript>
    <div id="app"></div>
    <!-- built files will be auto injected -->
  </body>
</html>
""",
                    "src/main.js": """
import { createApp } from 'vue'
import App from './App.vue'

createApp(App).mount('#app')
""",
                    "src/App.vue": """
<template>
  <div id="app">
    <header class="app-header">
      <h1>{{ app_title }}</h1>
      <p>{{ app_description }}</p>
    </header>
    
    <main class="app-main">
      <!-- Main content will go here -->
      {{ content }}
    </main>
    
    <footer class="app-footer">
      <p>&copy; 2023 {{ app_title }}</p>
    </footer>
  </div>
</template>

<script>
export default {
  name: 'App',
  data() {
    return {
      app_title: '{{ app_title }}',
      app_description: '{{ app_description }}'
    }
  }
}
</script>

<style>
#app {
  font-family: Avenir, Helvetica, Arial, sans-serif;
  -webkit-font-smoothing: antialiased;
  -moz-osx-font-smoothing: grayscale;
  color: #2c3e50;
}

.app-header {
  background-color: #282c34;
  min-height: 20vh;
  display: flex;
  flex-direction: column;
  align-items: center;
  justify-content: center;
  font-size: calc(10px + 2vmin);
  color: white;
  padding: 1rem;
  text-align: center;
}

.app-main {
  padding: 2rem;
  max-width: 1200px;
  margin: 0 auto;
}

.app-footer {
  background-color: #f4f4f4;
  padding: 1rem;
  text-align: center;
  margin-top: 2rem;
}
</style>
"""
                }
            }
        )
    
    def register_template(self, template_id: str, template_name: str,
                         framework: str, config: Dict[str, Any]) -> None:
        """
        Register a template in the manager.
        
        Args:
            template_id: Unique identifier for the template.
            template_name: Display name for the template.
            framework: Framework to which the template belongs.
            config: Configuration for the template.
        """
        if template_id in self.templates:
            raise UIGenerationError(f"Template ID '{template_id}' already exists.")
        
        if framework not in self.template_dirs:
            raise UIGenerationError(f"Framework '{framework}' is not supported.")
        
        template = {
            "id": template_id,
            "name": template_name,
            "framework": framework,
            "config": config
        }
        
        self.templates[template_id] = template
    
    def get_template(self, template_id: str) -> Dict[str, Any]:
        """
        Get a template by ID.
        
        Args:
            template_id: ID of the template to get.
            
        Returns:
            Template configuration.
        """
        if template_id not in self.templates:
            raise UIGenerationError(f"Template ID '{template_id}' does not exist.")
        
        return self.templates[template_id]
    
    def get_templates_by_framework(self, framework: str) -> List[Dict[str, Any]]:
        """
        Get all templates for a framework.
        
        Args:
            framework: Framework to get templates for.
            
        Returns:
            List of templates for the framework.
        """
        if framework not in self.template_dirs:
            raise UIGenerationError(f"Framework '{framework}' is not supported.")
        
        return [template for template in self.templates.values() if template["framework"] == framework]
    
    def get_all_templates(self) -> Dict[str, Dict[str, Any]]:
        """
        Get all templates.
        
        Returns:
            Dictionary of all templates.
        """
        return self.templates
    
    def apply_template(self, template_id: str, output_dir: str,
                      context: Dict[str, Any]) -> str:
        """
        Apply a template to generate files.
        
        Args:
            template_id: ID of the template to apply.
            output_dir: Directory to output generated files.
            context: Context variables for template rendering.
            
        Returns:
            Path to the output directory.
        """
        template = self.get_template(template_id)
        
        # Create output directory if it doesn't exist
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        
        # Create Jinja2 environment
        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(self.template_dirs[template["framework"]]),
            autoescape=jinja2.select_autoescape(["html", "xml"])
        )
        
        # Process each file in the template
        for file_path, file_content in template["config"]["files"].items():
            # Create directory for file if it doesn't exist
            file_dir = os.path.dirname(os.path.join(output_dir, file_path))
            if file_dir and not os.path.exists(file_dir):
                os.makedirs(file_dir)
            
            # Render file content
            template_obj = jinja2.Template(file_content)
            rendered_content = template_obj.render(**context)
            
            # Write file
            with open(os.path.join(output_dir, file_path), "w") as f:
                f.write(rendered_content)
        
        return output_dir
    
    def create_custom_template(self, template_id: str, template_name: str,
                              framework: str, files: Dict[str, str],
                              description: str = "") -> None:
        """
        Create a custom template.
        
        Args:
            template_id: Unique identifier for the template.
            template_name: Display name for the template.
            framework: Framework to which the template belongs.
            files: Dictionary of file paths and contents.
            description: Description of the template.
        """
        config = {
            "description": description,
            "files": files
        }
        
        self.register_template(template_id, template_name, framework, config)
    
    def save_template(self, template_id: str) -> None:
        """
        Save a template to disk.
        
        Args:
            template_id: ID of the template to save.
        """
        template = self.get_template(template_id)
        
        # Create template directory
        template_dir = os.path.join(self.template_dirs[template["framework"]], template_id)
        if not os.path.exists(template_dir):
            os.makedirs(template_dir)
        
        # Save template metadata
        metadata = {
            "id": template["id"],
            "name": template["name"],
            "framework": template["framework"],
            "description": template["config"]["description"]
        }
        
        with open(os.path.join(template_dir, "metadata.json"), "w") as f:
            json.dump(metadata, f, indent=2)
        
        # Save template files
        for file_path, file_content in template["config"]["files"].items():
            # Create directory for file if it doesn't exist
            file_dir = os.path.dirname(os.path.join(template_dir, "files", file_path))
            if file_dir and not os.path.exists(file_dir):
                os.makedirs(file_dir)
            
            # Write file
            with open(os.path.join(template_dir, "files", file_path), "w") as f:
                f.write(file_content)
    
    def load_template_from_disk(self, framework: str, template_id: str) -> None:
        """
        Load a template from disk.
        
        Args:
            framework: Framework to which the template belongs.
            template_id: ID of the template to load.
        """
        template_dir = os.path.join(self.template_dirs[framework], template_id)
        
        if not os.path.exists(template_dir):
            raise UIGenerationError(f"Template directory '{template_dir}' does not exist.")
        
        # Load template metadata
        with open(os.path.join(template_dir, "metadata.json"), "r") as f:
            metadata = json.load(f)
        
        # Load template files
        files = {}
        files_dir = os.path.join(template_dir, "files")
        
        for root, _, filenames in os.walk(files_dir):
            for filename in filenames:
                file_path = os.path.join(root, filename)
                relative_path = os.path.relpath(file_path, files_dir)
                
                with open(file_path, "r") as f:
                    files[relative_path] = f.read()
        
        # Register template
        self.register_template(
            metadata["id"],
            metadata["name"],
            metadata["framework"],
            {
                "description": metadata["description"],
                "files": files
            }
        )
    
    def export_template(self, template_id: str, output_path: str) -> str:
        """
        Export a template to a zip file.
        
        Args:
            template_id: ID of the template to export.
            output_path: Path to save the exported template.
            
        Returns:
            Path to the exported template.
        """
        import zipfile
        
        template = self.get_template(template_id)
        
        # Create temporary directory
        with tempfile.TemporaryDirectory() as temp_dir:
            # Save template metadata
            metadata = {
                "id": template["id"],
                "name": template["name"],
                "framework": template["framework"],
                "description": template["config"]["description"]
            }
            
            with open(os.path.join(temp_dir, "metadata.json"), "w") as f:
                json.dump(metadata, f, indent=2)
            
            # Save template files
            for file_path, file_content in template["config"]["files"].items():
                # Create directory for file if it doesn't exist
                file_dir = os.path.dirname(os.path.join(temp_dir, "files", file_path))
                if file_dir and not os.path.exists(file_dir):
                    os.makedirs(file_dir)
                
                # Write file
                with open(os.path.join(temp_dir, "files", file_path), "w") as f:
                    f.write(file_content)
            
            # Create zip file
            zip_path = output_path if output_path.endswith(".zip") else f"{output_path}.zip"
            with zipfile.ZipFile(zip_path, "w") as zipf:
                for root, _, files in os.walk(temp_dir):
                    for file in files:
                        file_path = os.path.join(root, file)
                        arcname = os.path.relpath(file_path, temp_dir)
                        zipf.write(file_path, arcname)
        
        return zip_path
    
    def import_template(self, zip_path: str) -> str:
        """
        Import a template from a zip file.
        
        Args:
            zip_path: Path to the zip file.
            
        Returns:
            ID of the imported template.
        """
        import zipfile
        
        # Create temporary directory
        with tempfile.TemporaryDirectory() as temp_dir:
            # Extract zip file
            with zipfile.ZipFile(zip_path, "r") as zipf:
                zipf.extractall(temp_dir)
            
            # Load template metadata
            with open(os.path.join(temp_dir, "metadata.json"), "r") as f:
                metadata = json.load(f)
            
            # Load template files
            files = {}
            files_dir = os.path.join(temp_dir, "files")
            
            for root, _, filenames in os.walk(files_dir):
                for filename in filenames:
                    file_path = os.path.join(root, filename)
                    relative_path = os.path.relpath(file_path, files_dir)
                    
                    with open(file_path, "r") as f:
                        files[relative_path] = f.read()
            
            # Register template
            self.register_template(
                metadata["id"],
                metadata["name"],
                metadata["framework"],
                {
                    "description": metadata["description"],
                    "files": files
                }
            )
            
            return metadata["id"]



"""
Data profiler for SBYB EDA.

This module provides functionality to automatically profile datasets
and generate comprehensive reports about their characteristics.
"""

from typing import Any, Dict, List, Optional, Union, Tuple
import os
import json
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from pathlib import Path
import datetime

from sbyb.core.base import SBYBComponent
from sbyb.core.config import Config


class DataProfiler(SBYBComponent):
    """
    Data profiler for exploratory data analysis.
    
    This component provides functionality to automatically profile datasets
    and generate comprehensive reports about their characteristics.
    """
    
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """
        Initialize the data profiler.
        
        Args:
            config: Configuration dictionary for the data profiler.
        """
        super().__init__(config)
        self.profile_results = {}
    
    def profile(self, data: pd.DataFrame, target_column: Optional[str] = None,
               include_correlations: bool = True,
               include_missing_analysis: bool = True,
               include_cardinality_analysis: bool = True,
               include_statistical_analysis: bool = True,
               include_distribution_analysis: bool = True) -> Dict[str, Any]:
        """
        Profile a dataset and generate a comprehensive report.
        
        Args:
            data: DataFrame to profile.
            target_column: Name of the target column, if any.
            include_correlations: Whether to include correlation analysis.
            include_missing_analysis: Whether to include missing value analysis.
            include_cardinality_analysis: Whether to include cardinality analysis.
            include_statistical_analysis: Whether to include statistical analysis.
            include_distribution_analysis: Whether to include distribution analysis.
            
        Returns:
            Dictionary containing the profile results.
        """
        if not isinstance(data, pd.DataFrame):
            raise TypeError("Data must be a pandas DataFrame")
        
        # Create a copy to avoid modifying the original
        df = data.copy()
        
        # Initialize profile results
        profile = {
            "dataset_info": self._get_dataset_info(df),
            "columns": self._get_column_info(df),
            "summary_statistics": self._get_summary_statistics(df) if include_statistical_analysis else {},
        }
        
        # Add missing value analysis
        if include_missing_analysis:
            profile["missing_analysis"] = self._analyze_missing_values(df)
        
        # Add cardinality analysis
        if include_cardinality_analysis:
            profile["cardinality_analysis"] = self._analyze_cardinality(df)
        
        # Add correlation analysis
        if include_correlations:
            profile["correlation_analysis"] = self._analyze_correlations(df, target_column)
        
        # Add distribution analysis
        if include_distribution_analysis:
            profile["distribution_analysis"] = self._analyze_distributions(df)
        
        # Add target analysis if target column is provided
        if target_column is not None and target_column in df.columns:
            profile["target_analysis"] = self._analyze_target(df, target_column)
        
        # Store profile results
        self.profile_results = profile
        
        return profile
    
    def _get_dataset_info(self, df: pd.DataFrame) -> Dict[str, Any]:
        """
        Get basic information about the dataset.
        
        Args:
            df: DataFrame to analyze.
            
        Returns:
            Dictionary containing dataset information.
        """
        return {
            "num_rows": len(df),
            "num_columns": len(df.columns),
            "memory_usage": df.memory_usage(deep=True).sum(),
            "memory_usage_formatted": f"{df.memory_usage(deep=True).sum() / (1024 * 1024):.2f} MB",
            "num_duplicate_rows": df.duplicated().sum(),
            "percent_duplicate_rows": f"{df.duplicated().mean() * 100:.2f}%",
            "column_types": {
                "numeric": len(df.select_dtypes(include=["number"]).columns),
                "categorical": len(df.select_dtypes(include=["object", "category"]).columns),
                "datetime": len(df.select_dtypes(include=["datetime"]).columns),
                "boolean": len(df.select_dtypes(include=["bool"]).columns),
                "other": len(df.columns) - len(df.select_dtypes(include=["number", "object", "category", "datetime", "bool"]).columns)
            },
            "profiling_timestamp": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        }
    
    def _get_column_info(self, df: pd.DataFrame) -> Dict[str, Dict[str, Any]]:
        """
        Get information about each column in the dataset.
        
        Args:
            df: DataFrame to analyze.
            
        Returns:
            Dictionary containing column information.
        """
        column_info = {}
        
        for column in df.columns:
            col_type = str(df[column].dtype)
            
            # Determine column category
            if pd.api.types.is_numeric_dtype(df[column]):
                category = "numeric"
            elif pd.api.types.is_datetime64_dtype(df[column]):
                category = "datetime"
            elif pd.api.types.is_bool_dtype(df[column]):
                category = "boolean"
            elif pd.api.types.is_categorical_dtype(df[column]):
                category = "categorical"
            elif pd.api.types.is_object_dtype(df[column]):
                # Check if it might be a datetime
                if df[column].dropna().apply(lambda x: isinstance(x, str) and self._is_date_string(x)).all():
                    category = "potential_datetime"
                else:
                    category = "categorical"
            else:
                category = "other"
            
            # Get basic column stats
            missing_count = df[column].isna().sum()
            missing_percent = missing_count / len(df) * 100
            
            column_info[column] = {
                "type": col_type,
                "category": category,
                "missing_count": missing_count,
                "missing_percent": f"{missing_percent:.2f}%",
                "unique_count": df[column].nunique(),
                "unique_percent": f"{df[column].nunique() / len(df) * 100:.2f}%",
                "memory_usage": df[column].memory_usage(deep=True),
                "memory_usage_formatted": f"{df[column].memory_usage(deep=True) / (1024 * 1024):.4f} MB"
            }
            
            # Add category-specific information
            if category == "numeric":
                column_info[column].update({
                    "min": df[column].min(),
                    "max": df[column].max(),
                    "mean": df[column].mean(),
                    "median": df[column].median(),
                    "std": df[column].std(),
                    "skewness": df[column].skew(),
                    "kurtosis": df[column].kurtosis(),
                    "zeros_count": (df[column] == 0).sum(),
                    "zeros_percent": f"{(df[column] == 0).mean() * 100:.2f}%",
                    "negative_count": (df[column] < 0).sum(),
                    "negative_percent": f"{(df[column] < 0).mean() * 100:.2f}%"
                })
            elif category in ["categorical", "potential_datetime"]:
                # Get top values and their frequencies
                value_counts = df[column].value_counts().head(10).to_dict()
                column_info[column]["top_values"] = value_counts
                
                # Check if it might be a boolean
                if df[column].nunique() == 2:
                    column_info[column]["potential_boolean"] = True
                    column_info[column]["boolean_values"] = df[column].dropna().unique().tolist()
            elif category == "datetime":
                column_info[column].update({
                    "min": df[column].min(),
                    "max": df[column].max(),
                    "range_days": (df[column].max() - df[column].min()).days,
                    "year_distribution": df[column].dt.year.value_counts().to_dict()
                })
            elif category == "boolean":
                column_info[column].update({
                    "true_count": df[column].sum(),
                    "true_percent": f"{df[column].mean() * 100:.2f}%",
                    "false_count": (~df[column]).sum(),
                    "false_percent": f"{(~df[column]).mean() * 100:.2f}%"
                })
        
        return column_info
    
    def _get_summary_statistics(self, df: pd.DataFrame) -> Dict[str, Any]:
        """
        Get summary statistics for the dataset.
        
        Args:
            df: DataFrame to analyze.
            
        Returns:
            Dictionary containing summary statistics.
        """
        # Get basic summary statistics
        summary = {}
        
        # Numeric columns
        numeric_df = df.select_dtypes(include=["number"])
        if not numeric_df.empty:
            summary["numeric"] = numeric_df.describe().to_dict()
        
        # Categorical columns
        categorical_df = df.select_dtypes(include=["object", "category"])
        if not categorical_df.empty:
            cat_summary = {}
            for column in categorical_df.columns:
                cat_summary[column] = {
                    "count": categorical_df[column].count(),
                    "unique": categorical_df[column].nunique(),
                    "top": categorical_df[column].value_counts().index[0] if not categorical_df[column].value_counts().empty else None,
                    "freq": categorical_df[column].value_counts().iloc[0] if not categorical_df[column].value_counts().empty else 0,
                    "top_5": categorical_df[column].value_counts().head(5).to_dict()
                }
            summary["categorical"] = cat_summary
        
        # Datetime columns
        datetime_df = df.select_dtypes(include=["datetime"])
        if not datetime_df.empty:
            dt_summary = {}
            for column in datetime_df.columns:
                dt_summary[column] = {
                    "count": datetime_df[column].count(),
                    "min": datetime_df[column].min(),
                    "max": datetime_df[column].max(),
                    "range_days": (datetime_df[column].max() - datetime_df[column].min()).days
                }
            summary["datetime"] = dt_summary
        
        return summary
    
    def _analyze_missing_values(self, df: pd.DataFrame) -> Dict[str, Any]:
        """
        Analyze missing values in the dataset.
        
        Args:
            df: DataFrame to analyze.
            
        Returns:
            Dictionary containing missing value analysis.
        """
        # Get missing value counts and percentages
        missing_counts = df.isna().sum()
        missing_percents = df.isna().mean() * 100
        
        # Create missing value summary
        missing_summary = {
            "total_missing_values": df.isna().sum().sum(),
            "total_missing_percent": f"{df.isna().mean().mean() * 100:.2f}%",
            "columns_with_missing": missing_counts[missing_counts > 0].shape[0],
            "columns_without_missing": missing_counts[missing_counts == 0].shape[0],
            "column_missing_counts": missing_counts[missing_counts > 0].to_dict(),
            "column_missing_percents": {col: f"{percent:.2f}%" for col, percent in missing_percents[missing_percents > 0].to_dict().items()},
            "rows_with_missing": df.isna().any(axis=1).sum(),
            "rows_with_missing_percent": f"{df.isna().any(axis=1).mean() * 100:.2f}%",
            "rows_without_missing": (~df.isna().any(axis=1)).sum(),
            "rows_without_missing_percent": f"{(~df.isna().any(axis=1)).mean() * 100:.2f}%"
        }
        
        # Add missing patterns
        missing_patterns = df.isna().groupby(list(df.columns)).size().reset_index(name="count")
        if not missing_patterns.empty:
            # Convert to more readable format
            patterns = []
            for _, row in missing_patterns.iterrows():
                pattern = {col: "Missing" if row[col] else "Present" for col in df.columns}
                pattern["count"] = row["count"]
                pattern["percent"] = f"{row['count'] / len(df) * 100:.2f}%"
                patterns.append(pattern)
            
            missing_summary["missing_patterns"] = patterns
        
        return missing_summary
    
    def _analyze_cardinality(self, df: pd.DataFrame) -> Dict[str, Any]:
        """
        Analyze cardinality of columns in the dataset.
        
        Args:
            df: DataFrame to analyze.
            
        Returns:
            Dictionary containing cardinality analysis.
        """
        # Get unique counts and percentages
        unique_counts = df.nunique()
        unique_percents = unique_counts / len(df) * 100
        
        # Categorize columns by cardinality
        high_cardinality = unique_percents[unique_percents > 90].index.tolist()
        medium_cardinality = unique_percents[(unique_percents > 50) & (unique_percents <= 90)].index.tolist()
        low_cardinality = unique_percents[(unique_percents > 10) & (unique_percents <= 50)].index.tolist()
        very_low_cardinality = unique_percents[unique_percents <= 10].index.tolist()
        
        # Create cardinality summary
        cardinality_summary = {
            "column_unique_counts": unique_counts.to_dict(),
            "column_unique_percents": {col: f"{percent:.2f}%" for col, percent in unique_percents.to_dict().items()},
            "high_cardinality_columns": high_cardinality,
            "medium_cardinality_columns": medium_cardinality,
            "low_cardinality_columns": low_cardinality,
            "very_low_cardinality_columns": very_low_cardinality,
            "potential_id_columns": high_cardinality
        }
        
        # Identify potential categorical columns that should be encoded
        categorical_df = df.select_dtypes(include=["object", "category"])
        if not categorical_df.empty:
            potential_encoded = []
            for column in categorical_df.columns:
                if unique_counts[column] <= 50:  # Reasonable number of categories
                    potential_encoded.append(column)
            
            cardinality_summary["potential_encoded_columns"] = potential_encoded
        
        # Identify potential continuous columns that should be binned
        numeric_df = df.select_dtypes(include=["number"])
        if not numeric_df.empty:
            potential_binned = []
            for column in numeric_df.columns:
                if unique_counts[column] > 100:  # Many unique values
                    potential_binned.append(column)
            
            cardinality_summary["potential_binned_columns"] = potential_binned
        
        return cardinality_summary
    
    def _analyze_correlations(self, df: pd.DataFrame, target_column: Optional[str] = None) -> Dict[str, Any]:
        """
        Analyze correlations between columns in the dataset.
        
        Args:
            df: DataFrame to analyze.
            target_column: Name of the target column, if any.
            
        Returns:
            Dictionary containing correlation analysis.
        """
        # Get numeric columns for correlation analysis
        numeric_df = df.select_dtypes(include=["number"])
        
        if numeric_df.empty:
            return {"message": "No numeric columns found for correlation analysis"}
        
        # Calculate correlation matrix
        corr_matrix = numeric_df.corr().round(3)
        
        # Create correlation summary
        correlation_summary = {
            "correlation_matrix": corr_matrix.to_dict()
        }
        
        # Find highly correlated pairs
        corr_pairs = []
        for i in range(len(corr_matrix.columns)):
            for j in range(i+1, len(corr_matrix.columns)):
                col1 = corr_matrix.columns[i]
                col2 = corr_matrix.columns[j]
                corr_value = corr_matrix.iloc[i, j]
                
                if abs(corr_value) > 0.7:  # High correlation threshold
                    corr_pairs.append({
                        "column1": col1,
                        "column2": col2,
                        "correlation": corr_value,
                        "strength": "strong positive" if corr_value > 0.7 else "strong negative"
                    })
        
        correlation_summary["highly_correlated_pairs"] = corr_pairs
        
        # Add target correlation if target column is provided
        if target_column is not None and target_column in numeric_df.columns:
            target_corrs = corr_matrix[target_column].drop(target_column).sort_values(ascending=False)
            
            correlation_summary["target_correlations"] = target_corrs.to_dict()
            
            # Identify top features by correlation
            correlation_summary["top_positive_features"] = target_corrs[target_corrs > 0].head(5).to_dict()
            correlation_summary["top_negative_features"] = target_corrs[target_corrs < 0].sort_values().head(5).to_dict()
        
        return correlation_summary
    
    def _analyze_distributions(self, df: pd.DataFrame) -> Dict[str, Any]:
        """
        Analyze distributions of columns in the dataset.
        
        Args:
            df: DataFrame to analyze.
            
        Returns:
            Dictionary containing distribution analysis.
        """
        distribution_summary = {}
        
        # Analyze numeric columns
        numeric_df = df.select_dtypes(include=["number"])
        if not numeric_df.empty:
            numeric_summary = {}
            
            for column in numeric_df.columns:
                # Skip columns with too many missing values
                if numeric_df[column].isna().mean() > 0.5:
                    continue
                
                # Calculate distribution statistics
                skewness = numeric_df[column].skew()
                kurtosis = numeric_df[column].kurtosis()
                
                # Determine distribution type
                dist_type = "unknown"
                if abs(skewness) < 0.5 and abs(kurtosis) < 0.5:
                    dist_type = "normal"
                elif skewness > 1:
                    dist_type = "right-skewed"
                elif skewness < -1:
                    dist_type = "left-skewed"
                
                # Check for outliers
                q1 = numeric_df[column].quantile(0.25)
                q3 = numeric_df[column].quantile(0.75)
                iqr = q3 - q1
                lower_bound = q1 - 1.5 * iqr
                upper_bound = q3 + 1.5 * iqr
                outliers = numeric_df[(numeric_df[column] < lower_bound) | (numeric_df[column] > upper_bound)][column]
                
                numeric_summary[column] = {
                    "skewness": skewness,
                    "kurtosis": kurtosis,
                    "distribution_type": dist_type,
                    "quartiles": {
                        "q1": q1,
                        "q2": numeric_df[column].median(),
                        "q3": q3,
                        "iqr": iqr
                    },
                    "outliers": {
                        "count": len(outliers),
                        "percent": f"{len(outliers) / len(numeric_df) * 100:.2f}%",
                        "min": outliers.min() if not outliers.empty else None,
                        "max": outliers.max() if not outliers.empty else None
                    }
                }
                
                # Suggest transformations for skewed distributions
                if abs(skewness) > 1:
                    transformations = []
                    if skewness > 1:  # Right-skewed
                        transformations.extend(["log", "sqrt", "box-cox"])
                    else:  # Left-skewed
                        transformations.extend(["square", "cube", "exponential"])
                    
                    numeric_summary[column]["suggested_transformations"] = transformations
            
            distribution_summary["numeric"] = numeric_summary
        
        # Analyze categorical columns
        categorical_df = df.select_dtypes(include=["object", "category"])
        if not categorical_df.empty:
            categorical_summary = {}
            
            for column in categorical_df.columns:
                # Skip columns with too many missing values
                if categorical_df[column].isna().mean() > 0.5:
                    continue
                
                # Calculate value counts
                value_counts = categorical_df[column].value_counts()
                
                # Check for imbalance
                if len(value_counts) > 1:
                    imbalance_ratio = value_counts.max() / value_counts.min()
                    is_imbalanced = imbalance_ratio > 10  # Arbitrary threshold
                else:
                    imbalance_ratio = float('inf')
                    is_imbalanced = True
                
                categorical_summary[column] = {
                    "value_counts": value_counts.head(10).to_dict(),
                    "imbalance_ratio": imbalance_ratio,
                    "is_imbalanced": is_imbalanced,
                    "entropy": self._calculate_entropy(value_counts / value_counts.sum()),
                    "suggested_encoding": self._suggest_encoding(value_counts)
                }
            
            distribution_summary["categorical"] = categorical_summary
        
        return distribution_summary
    
    def _analyze_target(self, df: pd.DataFrame, target_column: str) -> Dict[str, Any]:
        """
        Analyze the target column.
        
        Args:
            df: DataFrame to analyze.
            target_column: Name of the target column.
            
        Returns:
            Dictionary containing target analysis.
        """
        if target_column not in df.columns:
            return {"error": f"Target column '{target_column}' not found in the dataset"}
        
        target_series = df[target_column]
        target_type = str(target_series.dtype)
        
        # Determine target category
        if pd.api.types.is_numeric_dtype(target_series):
            category = "numeric"
        elif pd.api.types.is_datetime64_dtype(target_series):
            category = "datetime"
        elif pd.api.types.is_bool_dtype(target_series):
            category = "boolean"
        elif pd.api.types.is_categorical_dtype(target_series) or pd.api.types.is_object_dtype(target_series):
            category = "categorical"
        else:
            category = "other"
        
        # Initialize target analysis
        target_analysis = {
            "column": target_column,
            "type": target_type,
            "category": category,
            "missing_count": target_series.isna().sum(),
            "missing_percent": f"{target_series.isna().mean() * 100:.2f}%"
        }
        
        # Add category-specific analysis
        if category == "numeric":
            # Basic statistics
            target_analysis.update({
                "min": target_series.min(),
                "max": target_series.max(),
                "mean": target_series.mean(),
                "median": target_series.median(),
                "std": target_series.std(),
                "skewness": target_series.skew(),
                "kurtosis": target_series.kurtosis()
            })
            
            # Distribution analysis
            q1 = target_series.quantile(0.25)
            q3 = target_series.quantile(0.75)
            iqr = q3 - q1
            
            target_analysis["distribution"] = {
                "quartiles": {
                    "q1": q1,
                    "q2": target_series.median(),
                    "q3": q3,
                    "iqr": iqr
                },
                "distribution_type": "unknown"
            }
            
            # Determine distribution type
            skewness = target_series.skew()
            if abs(skewness) < 0.5:
                target_analysis["distribution"]["distribution_type"] = "normal"
            elif skewness > 1:
                target_analysis["distribution"]["distribution_type"] = "right-skewed"
            elif skewness < -1:
                target_analysis["distribution"]["distribution_type"] = "left-skewed"
            
            # Suggest task type
            target_analysis["suggested_task"] = "regression"
            
            # Suggest transformations for skewed distributions
            if abs(skewness) > 1:
                transformations = []
                if skewness > 1:  # Right-skewed
                    transformations.extend(["log", "sqrt", "box-cox"])
                else:  # Left-skewed
                    transformations.extend(["square", "cube", "exponential"])
                
                target_analysis["suggested_transformations"] = transformations
        
        elif category == "categorical":
            # Value counts
            value_counts = target_series.value_counts()
            
            target_analysis.update({
                "unique_count": target_series.nunique(),
                "unique_values": target_series.unique().tolist(),
                "value_counts": value_counts.to_dict()
            })
            
            # Check for imbalance
            if len(value_counts) > 1:
                imbalance_ratio = value_counts.max() / value_counts.min()
                is_imbalanced = imbalance_ratio > 10  # Arbitrary threshold
            else:
                imbalance_ratio = float('inf')
                is_imbalanced = True
            
            target_analysis["imbalance"] = {
                "imbalance_ratio": imbalance_ratio,
                "is_imbalanced": is_imbalanced
            }
            
            # Suggest task type
            if target_series.nunique() == 2:
                target_analysis["suggested_task"] = "binary_classification"
            else:
                target_analysis["suggested_task"] = "multiclass_classification"
            
            # Suggest handling for imbalanced data
            if is_imbalanced:
                target_analysis["suggested_handling"] = [
                    "class_weights",
                    "oversampling",
                    "undersampling",
                    "SMOTE"
                ]
        
        elif category == "boolean":
            # Value counts
            value_counts = target_series.value_counts()
            
            target_analysis.update({
                "true_count": target_series.sum(),
                "true_percent": f"{target_series.mean() * 100:.2f}%",
                "false_count": (~target_series).sum(),
                "false_percent": f"{(~target_series).mean() * 100:.2f}%"
            })
            
            # Check for imbalance
            imbalance_ratio = max(target_series.mean(), 1 - target_series.mean()) / min(target_series.mean(), 1 - target_series.mean())
            is_imbalanced = imbalance_ratio > 3  # Arbitrary threshold
            
            target_analysis["imbalance"] = {
                "imbalance_ratio": imbalance_ratio,
                "is_imbalanced": is_imbalanced
            }
            
            # Suggest task type
            target_analysis["suggested_task"] = "binary_classification"
            
            # Suggest handling for imbalanced data
            if is_imbalanced:
                target_analysis["suggested_handling"] = [
                    "class_weights",
                    "oversampling",
                    "undersampling",
                    "SMOTE"
                ]
        
        elif category == "datetime":
            # Basic statistics
            target_analysis.update({
                "min": target_series.min(),
                "max": target_series.max(),
                "range_days": (target_series.max() - target_series.min()).days
            })
            
            # Suggest task type
            target_analysis["suggested_task"] = "time_series"
        
        return target_analysis
    
    def _calculate_entropy(self, probabilities: pd.Series) -> float:
        """
        Calculate entropy of a probability distribution.
        
        Args:
            probabilities: Series of probabilities.
            
        Returns:
            Entropy value.
        """
        return -np.sum(probabilities * np.log2(probabilities))
    
    def _suggest_encoding(self, value_counts: pd.Series) -> str:
        """
        Suggest encoding method for a categorical column.
        
        Args:
            value_counts: Series of value counts.
            
        Returns:
            Suggested encoding method.
        """
        num_categories = len(value_counts)
        
        if num_categories == 2:
            return "binary"
        elif num_categories <= 10:
            return "one-hot"
        elif num_categories <= 50:
            return "label"
        else:
            return "target"
    
    def _is_date_string(self, value: str) -> bool:
        """
        Check if a string might be a date.
        
        Args:
            value: String to check.
            
        Returns:
            True if the string might be a date, False otherwise.
        """
        import re
        
        # Common date patterns
        date_patterns = [
            r'\d{4}-\d{2}-\d{2}',  # YYYY-MM-DD
            r'\d{2}-\d{2}-\d{4}',  # DD-MM-YYYY or MM-DD-YYYY
            r'\d{2}/\d{2}/\d{4}',  # DD/MM/YYYY or MM/DD/YYYY
            r'\d{4}/\d{2}/\d{2}',  # YYYY/MM/DD
            r'\d{2}\.\d{2}\.\d{4}',  # DD.MM.YYYY or MM.DD.YYYY
            r'\d{4}\.\d{2}\.\d{2}'   # YYYY.MM.DD
        ]
        
        for pattern in date_patterns:
            if re.match(pattern, value):
                return True
        
        return False
    
    def generate_html_report(self, output_path: str) -> str:
        """
        Generate an HTML report from the profile results.
        
        Args:
            output_path: Path to save the HTML report.
            
        Returns:
            Path to the generated HTML report.
        """
        if not self.profile_results:
            raise ValueError("No profile results available. Run profile() first.")
        
        # Create HTML content
        html_content = f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>SBYB Data Profile Report</title>
            <style>
                body {{
                    font-family: Arial, sans-serif;
                    margin: 20px;
                    color: #333;
                }}
                h1, h2, h3, h4 {{
                    color: #2c3e50;
                }}
                .container {{
                    max-width: 1200px;
                    margin: 0 auto;
                }}
                .section {{
                    margin-bottom: 30px;
                    padding: 20px;
                    background-color: #f9f9f9;
                    border-radius: 5px;
                    box-shadow: 0 2px 5px rgba(0,0,0,0.1);
                }}
                table {{
                    width: 100%;
                    border-collapse: collapse;
                    margin-bottom: 20px;
                }}
                th, td {{
                    padding: 12px 15px;
                    text-align: left;
                    border-bottom: 1px solid #ddd;
                }}
                th {{
                    background-color: #f2f2f2;
                }}
                tr:hover {{
                    background-color: #f5f5f5;
                }}
                .highlight {{
                    background-color: #ffffcc;
                }}
                .warning {{
                    color: #e74c3c;
                }}
                .success {{
                    color: #2ecc71;
                }}
                .info {{
                    color: #3498db;
                }}
            </style>
        </head>
        <body>
            <div class="container">
                <h1>SBYB Data Profile Report</h1>
                <p>Generated on {self.profile_results.get('dataset_info', {}).get('profiling_timestamp', datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))}</p>
                
                <div class="section">
                    <h2>Dataset Overview</h2>
                    <table>
                        <tr>
                            <th>Metric</th>
                            <th>Value</th>
                        </tr>
                        <tr>
                            <td>Number of Rows</td>
                            <td>{self.profile_results.get('dataset_info', {}).get('num_rows', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Number of Columns</td>
                            <td>{self.profile_results.get('dataset_info', {}).get('num_columns', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Memory Usage</td>
                            <td>{self.profile_results.get('dataset_info', {}).get('memory_usage_formatted', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Duplicate Rows</td>
                            <td>{self.profile_results.get('dataset_info', {}).get('num_duplicate_rows', 'N/A')} ({self.profile_results.get('dataset_info', {}).get('percent_duplicate_rows', 'N/A')})</td>
                        </tr>
                    </table>
                    
                    <h3>Column Types</h3>
                    <table>
                        <tr>
                            <th>Type</th>
                            <th>Count</th>
                        </tr>
                        <tr>
                            <td>Numeric</td>
                            <td>{self.profile_results.get('dataset_info', {}).get('column_types', {}).get('numeric', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Categorical</td>
                            <td>{self.profile_results.get('dataset_info', {}).get('column_types', {}).get('categorical', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Datetime</td>
                            <td>{self.profile_results.get('dataset_info', {}).get('column_types', {}).get('datetime', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Boolean</td>
                            <td>{self.profile_results.get('dataset_info', {}).get('column_types', {}).get('boolean', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Other</td>
                            <td>{self.profile_results.get('dataset_info', {}).get('column_types', {}).get('other', 'N/A')}</td>
                        </tr>
                    </table>
                </div>
        """
        
        # Add missing values section
        if 'missing_analysis' in self.profile_results:
            missing = self.profile_results['missing_analysis']
            html_content += f"""
                <div class="section">
                    <h2>Missing Values</h2>
                    <table>
                        <tr>
                            <th>Metric</th>
                            <th>Value</th>
                        </tr>
                        <tr>
                            <td>Total Missing Values</td>
                            <td>{missing.get('total_missing_values', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Total Missing Percent</td>
                            <td>{missing.get('total_missing_percent', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Columns With Missing</td>
                            <td>{missing.get('columns_with_missing', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Columns Without Missing</td>
                            <td>{missing.get('columns_without_missing', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Rows With Missing</td>
                            <td>{missing.get('rows_with_missing', 'N/A')} ({missing.get('rows_with_missing_percent', 'N/A')})</td>
                        </tr>
                        <tr>
                            <td>Rows Without Missing</td>
                            <td>{missing.get('rows_without_missing', 'N/A')} ({missing.get('rows_without_missing_percent', 'N/A')})</td>
                        </tr>
                    </table>
                    
                    <h3>Columns with Missing Values</h3>
                    <table>
                        <tr>
                            <th>Column</th>
                            <th>Missing Count</th>
                            <th>Missing Percent</th>
                        </tr>
            """
            
            for col, count in missing.get('column_missing_counts', {}).items():
                percent = missing.get('column_missing_percents', {}).get(col, 'N/A')
                html_content += f"""
                        <tr>
                            <td>{col}</td>
                            <td>{count}</td>
                            <td>{percent}</td>
                        </tr>
                """
            
            html_content += """
                    </table>
                </div>
            """
        
        # Add column details section
        if 'columns' in self.profile_results:
            html_content += """
                <div class="section">
                    <h2>Column Details</h2>
                    <table>
                        <tr>
                            <th>Column</th>
                            <th>Type</th>
                            <th>Category</th>
                            <th>Missing</th>
                            <th>Unique</th>
                            <th>Details</th>
                        </tr>
            """
            
            for col, info in self.profile_results['columns'].items():
                details = ""
                
                if info.get('category') == 'numeric':
                    details = f"Min: {info.get('min', 'N/A')}, Max: {info.get('max', 'N/A')}, Mean: {info.get('mean', 'N/A'):.2f}, Median: {info.get('median', 'N/A'):.2f}"
                elif info.get('category') in ['categorical', 'potential_datetime']:
                    top_values = info.get('top_values', {})
                    if top_values:
                        top_item = list(top_values.items())[0]
                        details = f"Most common: {top_item[0]} ({top_item[1]} occurrences)"
                elif info.get('category') == 'datetime':
                    details = f"Range: {info.get('min', 'N/A')} to {info.get('max', 'N/A')}"
                elif info.get('category') == 'boolean':
                    details = f"True: {info.get('true_percent', 'N/A')}, False: {info.get('false_percent', 'N/A')}"
                
                html_content += f"""
                        <tr>
                            <td>{col}</td>
                            <td>{info.get('type', 'N/A')}</td>
                            <td>{info.get('category', 'N/A')}</td>
                            <td>{info.get('missing_percent', 'N/A')}</td>
                            <td>{info.get('unique_count', 'N/A')} ({info.get('unique_percent', 'N/A')})</td>
                            <td>{details}</td>
                        </tr>
                """
            
            html_content += """
                    </table>
                </div>
            """
        
        # Add correlation section
        if 'correlation_analysis' in self.profile_results and 'highly_correlated_pairs' in self.profile_results['correlation_analysis']:
            corr_pairs = self.profile_results['correlation_analysis']['highly_correlated_pairs']
            
            if corr_pairs:
                html_content += """
                    <div class="section">
                        <h2>Correlations</h2>
                        <h3>Highly Correlated Pairs</h3>
                        <table>
                            <tr>
                                <th>Column 1</th>
                                <th>Column 2</th>
                                <th>Correlation</th>
                                <th>Strength</th>
                            </tr>
                """
                
                for pair in corr_pairs:
                    html_content += f"""
                            <tr>
                                <td>{pair.get('column1', 'N/A')}</td>
                                <td>{pair.get('column2', 'N/A')}</td>
                                <td>{pair.get('correlation', 'N/A')}</td>
                                <td>{pair.get('strength', 'N/A')}</td>
                            </tr>
                    """
                
                html_content += """
                        </table>
                    </div>
                """
        
        # Add target analysis section
        if 'target_analysis' in self.profile_results:
            target = self.profile_results['target_analysis']
            
            html_content += f"""
                <div class="section">
                    <h2>Target Analysis</h2>
                    <h3>Target: {target.get('column', 'N/A')}</h3>
                    <table>
                        <tr>
                            <th>Metric</th>
                            <th>Value</th>
                        </tr>
                        <tr>
                            <td>Type</td>
                            <td>{target.get('type', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Category</td>
                            <td>{target.get('category', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Missing</td>
                            <td>{target.get('missing_count', 'N/A')} ({target.get('missing_percent', 'N/A')})</td>
                        </tr>
                        <tr>
                            <td>Suggested Task</td>
                            <td>{target.get('suggested_task', 'N/A')}</td>
                        </tr>
            """
            
            if target.get('category') == 'numeric':
                html_content += f"""
                        <tr>
                            <td>Min</td>
                            <td>{target.get('min', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Max</td>
                            <td>{target.get('max', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Mean</td>
                            <td>{target.get('mean', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Median</td>
                            <td>{target.get('median', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Standard Deviation</td>
                            <td>{target.get('std', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Skewness</td>
                            <td>{target.get('skewness', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Distribution Type</td>
                            <td>{target.get('distribution', {}).get('distribution_type', 'N/A')}</td>
                        </tr>
                """
                
                if 'suggested_transformations' in target:
                    html_content += f"""
                        <tr>
                            <td>Suggested Transformations</td>
                            <td>{', '.join(target.get('suggested_transformations', []))}</td>
                        </tr>
                    """
            
            elif target.get('category') in ['categorical', 'boolean']:
                html_content += f"""
                        <tr>
                            <td>Unique Values</td>
                            <td>{target.get('unique_count', 'N/A')}</td>
                        </tr>
                        <tr>
                            <td>Imbalanced</td>
                            <td>{'Yes' if target.get('imbalance', {}).get('is_imbalanced', False) else 'No'}</td>
                        </tr>
                        <tr>
                            <td>Imbalance Ratio</td>
                            <td>{target.get('imbalance', {}).get('imbalance_ratio', 'N/A')}</td>
                        </tr>
                """
                
                if 'suggested_handling' in target:
                    html_content += f"""
                        <tr>
                            <td>Suggested Handling</td>
                            <td>{', '.join(target.get('suggested_handling', []))}</td>
                        </tr>
                    """
            
            html_content += """
                    </table>
                </div>
            """
        
        # Close HTML
        html_content += """
            </div>
        </body>
        </html>
        """
        
        # Write HTML to file
        with open(output_path, "w") as f:
            f.write(html_content)
        
        return output_path
    
    def generate_json_report(self, output_path: str) -> str:
        """
        Generate a JSON report from the profile results.
        
        Args:
            output_path: Path to save the JSON report.
            
        Returns:
            Path to the generated JSON report.
        """
        if not self.profile_results:
            raise ValueError("No profile results available. Run profile() first.")
        
        # Convert non-serializable objects to strings
        def json_serializable(obj):
            if isinstance(obj, (np.int64, np.int32, np.int16, np.int8)):
                return int(obj)
            elif isinstance(obj, (np.float64, np.float32, np.float16)):
                return float(obj)
            elif isinstance(obj, np.ndarray):
                return obj.tolist()
            elif isinstance(obj, pd.Series):
                return obj.to_dict()
            elif isinstance(obj, pd.DataFrame):
                return obj.to_dict(orient='records')
            elif isinstance(obj, (pd.Timestamp, datetime.datetime, datetime.date)):
                return str(obj)
            elif isinstance(obj, set):
                return list(obj)
            else:
                return str(obj)
        
        # Write JSON to file
        with open(output_path, "w") as f:
            json.dump(self.profile_results, f, default=json_serializable, indent=2)
        
        return output_path
    
    def plot_missing_values(self, output_path: Optional[str] = None) -> plt.Figure:
        """
        Plot missing values in the dataset.
        
        Args:
            output_path: Path to save the plot.
            
        Returns:
            Matplotlib figure.
        """
        if not self.profile_results or 'missing_analysis' not in self.profile_results:
            raise ValueError("No missing value analysis available. Run profile() first.")
        
        missing = self.profile_results['missing_analysis']
        
        # Create figure
        fig, ax = plt.subplots(figsize=(12, 8))
        
        # Get columns with missing values
        columns = list(missing.get('column_missing_counts', {}).keys())
        counts = list(missing.get('column_missing_counts', {}).values())
        
        # Sort by missing count
        sorted_indices = np.argsort(counts)
        columns = [columns[i] for i in sorted_indices]
        counts = [counts[i] for i in sorted_indices]
        
        # Plot
        ax.barh(columns, counts, color='skyblue')
        ax.set_xlabel('Missing Count')
        ax.set_ylabel('Column')
        ax.set_title('Missing Values by Column')
        ax.grid(axis='x', linestyle='--', alpha=0.7)
        
        # Add percentages
        for i, count in enumerate(counts):
            percent = missing.get('column_missing_percents', {}).get(columns[i], 'N/A')
            ax.text(count + 0.5, i, percent, va='center')
        
        plt.tight_layout()
        
        # Save if output path is provided
        if output_path:
            plt.savefig(output_path, dpi=300, bbox_inches='tight')
        
        return fig
    
    def plot_correlations(self, output_path: Optional[str] = None) -> plt.Figure:
        """
        Plot correlation matrix.
        
        Args:
            output_path: Path to save the plot.
            
        Returns:
            Matplotlib figure.
        """
        if not self.profile_results or 'correlation_analysis' not in self.profile_results:
            raise ValueError("No correlation analysis available. Run profile() first.")
        
        if 'correlation_matrix' not in self.profile_results['correlation_analysis']:
            raise ValueError("No correlation matrix available.")
        
        # Get correlation matrix
        corr_matrix = self.profile_results['correlation_analysis']['correlation_matrix']
        
        # Convert to DataFrame
        corr_df = pd.DataFrame(corr_matrix)
        
        # Create figure
        fig, ax = plt.subplots(figsize=(12, 10))
        
        # Plot
        mask = np.triu(np.ones_like(corr_df, dtype=bool))
        cmap = sns.diverging_palette(230, 20, as_cmap=True)
        
        sns.heatmap(corr_df, mask=mask, cmap=cmap, vmax=1, vmin=-1, center=0,
                   square=True, linewidths=.5, cbar_kws={"shrink": .5}, annot=True, fmt=".2f")
        
        ax.set_title('Correlation Matrix')
        
        plt.tight_layout()
        
        # Save if output path is provided
        if output_path:
            plt.savefig(output_path, dpi=300, bbox_inches='tight')
        
        return fig
    
    def plot_distributions(self, columns: Optional[List[str]] = None, 
                          output_path: Optional[str] = None) -> plt.Figure:
        """
        Plot distributions of numeric columns.
        
        Args:
            columns: List of columns to plot. If None, plot all numeric columns.
            output_path: Path to save the plot.
            
        Returns:
            Matplotlib figure.
        """
        if not self.profile_results or 'columns' not in self.profile_results:
            raise ValueError("No column information available. Run profile() first.")
        
        # Get numeric columns
        numeric_columns = [col for col, info in self.profile_results['columns'].items() 
                          if info.get('category') == 'numeric']
        
        if not numeric_columns:
            raise ValueError("No numeric columns available.")
        
        # Filter columns if provided
        if columns:
            numeric_columns = [col for col in columns if col in numeric_columns]
            
            if not numeric_columns:
                raise ValueError("No valid numeric columns provided.")
        
        # Limit to 16 columns for readability
        if len(numeric_columns) > 16:
            numeric_columns = numeric_columns[:16]
        
        # Calculate grid dimensions
        n_cols = min(4, len(numeric_columns))
        n_rows = (len(numeric_columns) + n_cols - 1) // n_cols
        
        # Create figure
        fig, axes = plt.subplots(n_rows, n_cols, figsize=(15, 3 * n_rows))
        
        # Flatten axes for easy iteration
        if n_rows == 1 and n_cols == 1:
            axes = np.array([axes])
        axes = axes.flatten()
        
        # Plot distributions
        for i, col in enumerate(numeric_columns):
            if i < len(axes):
                # Get distribution info
                if 'distribution_analysis' in self.profile_results and 'numeric' in self.profile_results['distribution_analysis']:
                    dist_info = self.profile_results['distribution_analysis']['numeric'].get(col, {})
                    dist_type = dist_info.get('distribution_type', 'unknown')
                    skewness = dist_info.get('skewness', 0)
                else:
                    dist_type = 'unknown'
                    skewness = 0
                
                # Add title with distribution type
                title = f"{col}\n(Distribution: {dist_type}, Skewness: {skewness:.2f})"
                axes[i].set_title(title)
                
                # Add grid
                axes[i].grid(linestyle='--', alpha=0.7)
        
        # Hide unused subplots
        for i in range(len(numeric_columns), len(axes)):
            axes[i].set_visible(False)
        
        plt.tight_layout()
        
        # Save if output path is provided
        if output_path:
            plt.savefig(output_path, dpi=300, bbox_inches='tight')
        
        return fig
    
    def plot_categorical_distributions(self, columns: Optional[List[str]] = None,
                                      max_categories: int = 10,
                                      output_path: Optional[str] = None) -> plt.Figure:
        """
        Plot distributions of categorical columns.
        
        Args:
            columns: List of columns to plot. If None, plot all categorical columns.
            max_categories: Maximum number of categories to show per column.
            output_path: Path to save the plot.
            
        Returns:
            Matplotlib figure.
        """
        if not self.profile_results or 'columns' not in self.profile_results:
            raise ValueError("No column information available. Run profile() first.")
        
        # Get categorical columns
        categorical_columns = [col for col, info in self.profile_results['columns'].items() 
                              if info.get('category') in ['categorical', 'potential_datetime']]
        
        if not categorical_columns:
            raise ValueError("No categorical columns available.")
        
        # Filter columns if provided
        if columns:
            categorical_columns = [col for col in columns if col in categorical_columns]
            
            if not categorical_columns:
                raise ValueError("No valid categorical columns provided.")
        
        # Limit to 9 columns for readability
        if len(categorical_columns) > 9:
            categorical_columns = categorical_columns[:9]
        
        # Calculate grid dimensions
        n_cols = min(3, len(categorical_columns))
        n_rows = (len(categorical_columns) + n_cols - 1) // n_cols
        
        # Create figure
        fig, axes = plt.subplots(n_rows, n_cols, figsize=(15, 4 * n_rows))
        
        # Flatten axes for easy iteration
        if n_rows == 1 and n_cols == 1:
            axes = np.array([axes])
        axes = axes.flatten()
        
        # Plot distributions
        for i, col in enumerate(categorical_columns):
            if i < len(axes):
                # Get value counts
                if 'distribution_analysis' in self.profile_results and 'categorical' in self.profile_results['distribution_analysis']:
                    value_counts = self.profile_results['distribution_analysis']['categorical'].get(col, {}).get('value_counts', {})
                else:
                    value_counts = self.profile_results['columns'][col].get('top_values', {})
                
                # Convert to Series for easier handling
                value_counts = pd.Series(value_counts)
                
                # Limit categories
                if len(value_counts) > max_categories:
                    other_count = value_counts.iloc[max_categories:].sum()
                    value_counts = value_counts.iloc[:max_categories]
                    value_counts['Other'] = other_count
                
                # Plot
                value_counts.plot(kind='bar', ax=axes[i], color='skyblue')
                
                # Add title
                axes[i].set_title(col)
                
                # Rotate x-axis labels
                axes[i].set_xticklabels(axes[i].get_xticklabels(), rotation=45, ha='right')
                
                # Add grid
                axes[i].grid(axis='y', linestyle='--', alpha=0.7)
        
        # Hide unused subplots
        for i in range(len(categorical_columns), len(axes)):
            axes[i].set_visible(False)
        
        plt.tight_layout()
        
        # Save if output path is provided
        if output_path:
            plt.savefig(output_path, dpi=300, bbox_inches='tight')
        
        return fig



"""
Data preprocessing module for SBYB.

This module provides components for preprocessing data in the SBYB library.
"""

from sbyb.preprocessing.base import BasePreprocessor
from sbyb.preprocessing.cleaner import DataCleaner
from sbyb.preprocessing.imputer import MissingValueImputer
from sbyb.preprocessing.outlier import OutlierHandler
from sbyb.preprocessing.encoder import CategoricalEncoder
from sbyb.preprocessing.scaler import FeatureScaler
from sbyb.preprocessing.feature_engineering import FeatureEngineer
from sbyb.preprocessing.pipeline import PreprocessingPipeline

__all__ = [
    'BasePreprocessor',
    'DataCleaner',
    'MissingValueImputer',
    'OutlierHandler',
    'CategoricalEncoder',
    'FeatureScaler',
    'FeatureEngineer',
    'PreprocessingPipeline',
]



"""
API generator component for SBYB deployment.

This module provides functionality for automatically generating APIs
for machine learning models.
"""

from typing import Any, Dict, List, Optional, Union, Tuple
import os
import json
import shutil
import subprocess
import tempfile

from sbyb.core.base import SBYBComponent
from sbyb.core.exceptions import APIGenerationError


class APIGenerator(SBYBComponent):
    """
    API generator component.
    
    This component generates APIs for machine learning models.
    """
    
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """
        Initialize the API generator.
        
        Args:
            config: Configuration dictionary for the generator.
        """
        super().__init__(config)
    
    def generate_fastapi(self, model_path: str, output_dir: str, 
                        api_name: str = "model_api",
                        host: str = "0.0.0.0", 
                        port: int = 8000,
                        include_docs: bool = True,
                        include_preprocessing: bool = False,
                        preprocessor_path: Optional[str] = None) -> str:
        """
        Generate a FastAPI application for serving a model.
        
        Args:
            model_path: Path to the model file.
            output_dir: Directory to save the generated API.
            api_name: Name of the API.
            host: Host address to bind the server.
            port: Port to bind the server.
            include_docs: Whether to include API documentation.
            include_preprocessing: Whether to include preprocessing.
            preprocessor_path: Path to the preprocessor file.
            
        Returns:
            Path to the generated API.
        """
        # Create output directory if it doesn't exist
        os.makedirs(output_dir, exist_ok=True)
        
        # Create main.py
        main_py_content = self._generate_fastapi_main(
            model_path=model_path,
            api_name=api_name,
            host=host,
            port=port,
            include_docs=include_docs,
            include_preprocessing=include_preprocessing,
            preprocessor_path=preprocessor_path
        )
        
        with open(os.path.join(output_dir, "main.py"), "w") as f:
            f.write(main_py_content)
        
        # Create requirements.txt
        requirements_txt_content = self._generate_fastapi_requirements()
        
        with open(os.path.join(output_dir, "requirements.txt"), "w") as f:
            f.write(requirements_txt_content)
        
        # Create README.md
        readme_md_content = self._generate_fastapi_readme(
            api_name=api_name,
            host=host,
            port=port
        )
        
        with open(os.path.join(output_dir, "README.md"), "w") as f:
            f.write(readme_md_content)
        
        # Create Dockerfile
        dockerfile_content = self._generate_fastapi_dockerfile()
        
        with open(os.path.join(output_dir, "Dockerfile"), "w") as f:
            f.write(dockerfile_content)
        
        # Create .dockerignore
        dockerignore_content = self._generate_dockerignore()
        
        with open(os.path.join(output_dir, ".dockerignore"), "w") as f:
            f.write(dockerignore_content)
        
        # Copy model file to output directory
        model_filename = os.path.basename(model_path)
        shutil.copy(model_path, os.path.join(output_dir, model_filename))
        
        # Copy preprocessor file to output directory if provided
        if include_preprocessing and preprocessor_path is not None:
            preprocessor_filename = os.path.basename(preprocessor_path)
            shutil.copy(preprocessor_path, os.path.join(output_dir, preprocessor_filename))
        
        return output_dir
    
    def generate_flask(self, model_path: str, output_dir: str, 
                      api_name: str = "model_api",
                      host: str = "0.0.0.0", 
                      port: int = 5000,
                      include_preprocessing: bool = False,
                      preprocessor_path: Optional[str] = None) -> str:
        """
        Generate a Flask application for serving a model.
        
        Args:
            model_path: Path to the model file.
            output_dir: Directory to save the generated API.
            api_name: Name of the API.
            host: Host address to bind the server.
            port: Port to bind the server.
            include_preprocessing: Whether to include preprocessing.
            preprocessor_path: Path to the preprocessor file.
            
        Returns:
            Path to the generated API.
        """
        # Create output directory if it doesn't exist
        os.makedirs(output_dir, exist_ok=True)
        
        # Create app.py
        app_py_content = self._generate_flask_app(
            model_path=model_path,
            api_name=api_name,
            host=host,
            port=port,
            include_preprocessing=include_preprocessing,
            preprocessor_path=preprocessor_path
        )
        
        with open(os.path.join(output_dir, "app.py"), "w") as f:
            f.write(app_py_content)
        
        # Create requirements.txt
        requirements_txt_content = self._generate_flask_requirements()
        
        with open(os.path.join(output_dir, "requirements.txt"), "w") as f:
            f.write(requirements_txt_content)
        
        # Create README.md
        readme_md_content = self._generate_flask_readme(
            api_name=api_name,
            host=host,
            port=port
        )
        
        with open(os.path.join(output_dir, "README.md"), "w") as f:
            f.write(readme_md_content)
        
        # Create Dockerfile
        dockerfile_content = self._generate_flask_dockerfile()
        
        with open(os.path.join(output_dir, "Dockerfile"), "w") as f:
            f.write(dockerfile_content)
        
        # Create .dockerignore
        dockerignore_content = self._generate_dockerignore()
        
        with open(os.path.join(output_dir, ".dockerignore"), "w") as f:
            f.write(dockerignore_content)
        
        # Copy model file to output directory
        model_filename = os.path.basename(model_path)
        shutil.copy(model_path, os.path.join(output_dir, model_filename))
        
        # Copy preprocessor file to output directory if provided
        if include_preprocessing and preprocessor_path is not None:
            preprocessor_filename = os.path.basename(preprocessor_path)
            shutil.copy(preprocessor_path, os.path.join(output_dir, preprocessor_filename))
        
        return output_dir
    
    def generate_streamlit(self, model_path: str, output_dir: str, 
                          app_name: str = "Model Demo",
                          port: int = 8501,
                          include_preprocessing: bool = False,
                          preprocessor_path: Optional[str] = None,
                          include_visualization: bool = True) -> str:
        """
        Generate a Streamlit application for demonstrating a model.
        
        Args:
            model_path: Path to the model file.
            output_dir: Directory to save the generated application.
            app_name: Name of the application.
            port: Port to bind the server.
            include_preprocessing: Whether to include preprocessing.
            preprocessor_path: Path to the preprocessor file.
            include_visualization: Whether to include visualization.
            
        Returns:
            Path to the generated application.
        """
        # Create output directory if it doesn't exist
        os.makedirs(output_dir, exist_ok=True)
        
        # Create app.py
        app_py_content = self._generate_streamlit_app(
            model_path=model_path,
            app_name=app_name,
            include_preprocessing=include_preprocessing,
            preprocessor_path=preprocessor_path,
            include_visualization=include_visualization
        )
        
        with open(os.path.join(output_dir, "app.py"), "w") as f:
            f.write(app_py_content)
        
        # Create requirements.txt
        requirements_txt_content = self._generate_streamlit_requirements()
        
        with open(os.path.join(output_dir, "requirements.txt"), "w") as f:
            f.write(requirements_txt_content)
        
        # Create README.md
        readme_md_content = self._generate_streamlit_readme(
            app_name=app_name,
            port=port
        )
        
        with open(os.path.join(output_dir, "README.md"), "w") as f:
            f.write(readme_md_content)
        
        # Create Dockerfile
        dockerfile_content = self._generate_streamlit_dockerfile()
        
        with open(os.path.join(output_dir, "Dockerfile"), "w") as f:
            f.write(dockerfile_content)
        
        # Create .dockerignore
        dockerignore_content = self._generate_dockerignore()
        
        with open(os.path.join(output_dir, ".dockerignore"), "w") as f:
            f.write(dockerignore_content)
        
        # Copy model file to output directory
        model_filename = os.path.basename(model_path)
        shutil.copy(model_path, os.path.join(output_dir, model_filename))
        
        # Copy preprocessor file to output directory if provided
        if include_preprocessing and preprocessor_path is not None:
            preprocessor_filename = os.path.basename(preprocessor_path)
            shutil.copy(preprocessor_path, os.path.join(output_dir, preprocessor_filename))
        
        return output_dir
    
    def _generate_fastapi_main(self, model_path: str, api_name: str, host: str, port: int,
                              include_docs: bool, include_preprocessing: bool,
                              preprocessor_path: Optional[str]) -> str:
        """
        Generate FastAPI main.py content.
        
        Args:
            model_path: Path to the model file.
            api_name: Name of the API.
            host: Host address to bind the server.
            port: Port to bind the server.
            include_docs: Whether to include API documentation.
            include_preprocessing: Whether to include preprocessing.
            preprocessor_path: Path to the preprocessor file.
            
        Returns:
            Content of main.py.
        """
        model_filename = os.path.basename(model_path)
        preprocessor_filename = os.path.basename(preprocessor_path) if preprocessor_path else None
        
        content = f"""
import os
import pickle
import json
from typing import List, Dict, Any, Optional

import numpy as np
import pandas as pd
from fastapi import FastAPI, HTTPException
from pydantic import BaseModel
import uvicorn

# Create FastAPI app
app = FastAPI(
    title="{api_name}",
    description="API for serving machine learning model",
    version="1.0.0",
    {"openapi_url=None" if not include_docs else ""}
)

# Load model
try:
    with open("{model_filename}", "rb") as f:
        loaded_obj = pickle.load(f)
    
    # Check if the loaded object is a dictionary with model and metadata
    if isinstance(loaded_obj, dict) and 'model' in loaded_obj:
        model = loaded_obj['model']
        metadata = loaded_obj.get('metadata', {{}})
    else:
        # Assume the loaded object is the model itself
        model = loaded_obj
        metadata = {{}}
    
    print(f"Model loaded from {model_filename}")
except Exception as e:
    print(f"Error loading model: {{e}}")
    raise
"""
        
        if include_preprocessing and preprocessor_path:
            content += f"""
# Load preprocessor
try:
    with open("{preprocessor_filename}", "rb") as f:
        loaded_obj = pickle.load(f)
    
    # Check if the loaded object is a dictionary with preprocessor and metadata
    if isinstance(loaded_obj, dict) and 'preprocessor' in loaded_obj:
        preprocessor = loaded_obj['preprocessor']
    else:
        # Assume the loaded object is the preprocessor itself
        preprocessor = loaded_obj
    
    print(f"Preprocessor loaded from {preprocessor_filename}")
except Exception as e:
    print(f"Error loading preprocessor: {{e}}")
    preprocessor = None
"""
        else:
            content += """
# No preprocessor
preprocessor = None
"""
        
        content += """
# Define request and response models
class PredictionRequest(BaseModel):
    data: List[Dict[str, Any]]

class PredictionResponse(BaseModel):
    predictions: List[Any]

class ProbabilityResponse(BaseModel):
    probabilities: List[List[float]]

@app.post("/predict", response_model=PredictionResponse)
async def predict(request: PredictionRequest):
    try:
        # Convert input data to DataFrame
        data = pd.DataFrame(request.data)
        
        # Apply preprocessor if available
        if preprocessor is not None:
            try:
                data = preprocessor.transform(data)
            except Exception as e:
                raise HTTPException(status_code=400, detail=f"Error applying preprocessor: {str(e)}")
        
        # Make predictions
        predictions = model.predict(data)
        
        # Convert predictions to list
        if isinstance(predictions, np.ndarray):
            predictions = predictions.tolist()
        
        # Return predictions
        return {"predictions": predictions}
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))

@app.post("/predict_proba", response_model=ProbabilityResponse)
async def predict_proba(request: PredictionRequest):
    try:
        # Check if model supports probability predictions
        if not hasattr(model, 'predict_proba'):
            raise HTTPException(status_code=400, detail="Model does not support probability predictions")
        
        # Convert input data to DataFrame
        data = pd.DataFrame(request.data)
        
        # Apply preprocessor if available
        if preprocessor is not None:
            try:
                data = preprocessor.transform(data)
            except Exception as e:
                raise HTTPException(status_code=400, detail=f"Error applying preprocessor: {str(e)}")
        
        # Make probability predictions
        probabilities = model.predict_proba(data)
        
        # Convert predictions to list
        if isinstance(probabilities, np.ndarray):
            probabilities = probabilities.tolist()
        
        # Return probabilities
        return {"probabilities": probabilities}
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))

@app.get("/health")
async def health():
    return {"status": "healthy"}

@app.get("/metadata")
async def get_metadata():
    # Convert numpy arrays and other non-serializable objects to strings
    metadata_dict = {}
    for key, value in metadata.items():
        if isinstance(value, np.ndarray):
            metadata_dict[key] = value.tolist()
        elif isinstance(value, (pd.DataFrame, pd.Series)):
            metadata_dict[key] = value.to_dict()
        else:
            try:
                # Try to serialize, if it fails, convert to string
                json.dumps({key: value})
                metadata_dict[key] = value
            except (TypeError, OverflowError):
                metadata_dict[key] = str(value)
    
    return metadata_dict

if __name__ == "__main__":
    uvicorn.run("main:app", host="{host}", port={port}, reload=False)
"""
        
        return content
    
    def _generate_fastapi_requirements(self) -> str:
        """
        Generate FastAPI requirements.txt content.
        
        Returns:
            Content of requirements.txt.
        """
        return """
fastapi>=0.68.0
uvicorn>=0.15.0
pandas>=1.3.0
numpy>=1.20.0
scikit-learn>=0.24.0
pydantic>=1.8.0
python-multipart>=0.0.5
"""
    
    def _generate_fastapi_readme(self, api_name: str, host: str, port: int) -> str:
        """
        Generate FastAPI README.md content.
        
        Args:
            api_name: Name of the API.
            host: Host address to bind the server.
            port: Port to bind the server.
            
        Returns:
            Content of README.md.
        """
        return f"""
# {api_name}

This is a FastAPI application for serving a machine learning model.

## Installation

1. Clone this repository
2. Install dependencies:
   ```
   pip install -r requirements.txt
   ```

## Usage

### Running the API

```
python main.py
```

The API will be available at http://{host}:{port}

### API Endpoints

- `POST /predict`: Make predictions
  - Request body: `{{"data": [{{"feature1": value1, "feature2": value2, ...}}, ...]}}`
  - Response: `{{"predictions": [prediction1, prediction2, ...]}}`

- `POST /predict_proba`: Make probability predictions (if supported by the model)
  - Request body: `{{"data": [{{"feature1": value1, "feature2": value2, ...}}, ...]}}`
  - Response: `{{"probabilities": [[prob1_class1, prob1_class2, ...], [prob2_class1, prob2_class2, ...], ...]}}`

- `GET /health`: Health check
  - Response: `{{"status": "healthy"}}`

- `GET /metadata`: Get model metadata
  - Response: Model metadata

### API Documentation

API documentation is available at http://{host}:{port}/docs

## Docker

You can also run the API using Docker:

```
docker build -t {api_name.lower().replace(" ", "-")} .
docker run -p {port}:{port} {api_name.lower().replace(" ", "-")}
```
"""
    
    def _generate_fastapi_dockerfile(self) -> str:
        """
        Generate FastAPI Dockerfile content.
        
        Returns:
            Content of Dockerfile.
        """
        return """
FROM python:3.9-slim

WORKDIR /app

COPY requirements.txt .
RUN pip install --no-cache-dir -r requirements.txt

COPY . .

CMD ["python", "main.py"]
"""
    
    def _generate_flask_app(self, model_path: str, api_name: str, host: str, port: int,
                           include_preprocessing: bool, preprocessor_path: Optional[str]) -> str:
        """
        Generate Flask app.py content.
        
        Args:
            model_path: Path to the model file.
            api_name: Name of the API.
            host: Host address to bind the server.
            port: Port to bind the server.
            include_preprocessing: Whether to include preprocessing.
            preprocessor_path: Path to the preprocessor file.
            
        Returns:
            Content of app.py.
        """
        model_filename = os.path.basename(model_path)
        preprocessor_filename = os.path.basename(preprocessor_path) if preprocessor_path else None
        
        content = f"""
import os
import pickle
import json

import numpy as np
import pandas as pd
from flask import Flask, request, jsonify

# Create Flask app
app = Flask("{api_name}")

# Load model
try:
    with open("{model_filename}", "rb") as f:
        loaded_obj = pickle.load(f)
    
    # Check if the loaded object is a dictionary with model and metadata
    if isinstance(loaded_obj, dict) and 'model' in loaded_obj:
        model = loaded_obj['model']
        metadata = loaded_obj.get('metadata', {{}})
    else:
        # Assume the loaded object is the model itself
        model = loaded_obj
        metadata = {{}}
    
    print(f"Model loaded from {model_filename}")
except Exception as e:
    print(f"Error loading model: {{e}}")
    raise
"""
        
        if include_preprocessing and preprocessor_path:
            content += f"""
# Load preprocessor
try:
    with open("{preprocessor_filename}", "rb") as f:
        loaded_obj = pickle.load(f)
    
    # Check if the loaded object is a dictionary with preprocessor and metadata
    if isinstance(loaded_obj, dict) and 'preprocessor' in loaded_obj:
        preprocessor = loaded_obj['preprocessor']
    else:
        # Assume the loaded object is the preprocessor itself
        preprocessor = loaded_obj
    
    print(f"Preprocessor loaded from {preprocessor_filename}")
except Exception as e:
    print(f"Error loading preprocessor: {{e}}")
    preprocessor = None
"""
        else:
            content += """
# No preprocessor
preprocessor = None
"""
        
        content += """
@app.route('/predict', methods=['POST'])
def predict():
    try:
        # Get data from request
        data = request.json
        
        # Convert to DataFrame if it's a list of dictionaries
        if isinstance(data, list):
            data = pd.DataFrame(data)
        
        # Apply preprocessor if available
        if preprocessor is not None:
            try:
                data = preprocessor.transform(data)
            except Exception as e:
                return jsonify({'error': f"Error applying preprocessor: {str(e)}"}), 400
        
        # Make predictions
        predictions = model.predict(data)
        
        # Convert predictions to list
        if isinstance(predictions, np.ndarray):
            predictions = predictions.tolist()
        
        # Return predictions
        return jsonify({'predictions': predictions})
    except Exception as e:
        return jsonify({'error': str(e)}), 400

@app.route('/predict_proba', methods=['POST'])
def predict_proba():
    try:
        # Check if model supports probability predictions
        if not hasattr(model, 'predict_proba'):
            return jsonify({'error': 'Model does not support probability predictions'}), 400
        
        # Get data from request
        data = request.json
        
        # Convert to DataFrame if it's a list of dictionaries
        if isinstance(data, list):
            data = pd.DataFrame(data)
        
        # Apply preprocessor if available
        if preprocessor is not None:
            try:
                data = preprocessor.transform(data)
            except Exception as e:
                return jsonify({'error': f"Error applying preprocessor: {str(e)}"}), 400
        
        # Make probability predictions
        probabilities = model.predict_proba(data)
        
        # Convert predictions to list
        if isinstance(probabilities, np.ndarray):
            probabilities = probabilities.tolist()
        
        # Return probabilities
        return jsonify({'probabilities': probabilities})
    except Exception as e:
        return jsonify({'error': str(e)}), 400

@app.route('/health', methods=['GET'])
def health():
    return jsonify({'status': 'healthy'})

@app.route('/metadata', methods=['GET'])
def get_metadata():
    # Convert numpy arrays and other non-serializable objects to strings
    metadata_dict = {}
    for key, value in metadata.items():
        if isinstance(value, np.ndarray):
            metadata_dict[key] = value.tolist()
        elif isinstance(value, (pd.DataFrame, pd.Series)):
            metadata_dict[key] = value.to_dict()
        else:
            try:
                # Try to serialize, if it fails, convert to string
                json.dumps({key: value})
                metadata_dict[key] = value
            except (TypeError, OverflowError):
                metadata_dict[key] = str(value)
    
    return jsonify(metadata_dict)

if __name__ == '__main__':
    app.run(host='{host}', port={port})
"""
        
        return content
    
    def _generate_flask_requirements(self) -> str:
        """
        Generate Flask requirements.txt content.
        
        Returns:
            Content of requirements.txt.
        """
        return """
flask>=2.0.0
pandas>=1.3.0
numpy>=1.20.0
scikit-learn>=0.24.0
"""
    
    def _generate_flask_readme(self, api_name: str, host: str, port: int) -> str:
        """
        Generate Flask README.md content.
        
        Args:
            api_name: Name of the API.
            host: Host address to bind the server.
            port: Port to bind the server.
            
        Returns:
            Content of README.md.
        """
        return f"""
# {api_name}

This is a Flask application for serving a machine learning model.

## Installation

1. Clone this repository
2. Install dependencies:
   ```
   pip install -r requirements.txt
   ```

## Usage

### Running the API

```
python app.py
```

The API will be available at http://{host}:{port}

### API Endpoints

- `POST /predict`: Make predictions
  - Request body: `[{{"feature1": value1, "feature2": value2, ...}}, ...]`
  - Response: `{{"predictions": [prediction1, prediction2, ...]}}`

- `POST /predict_proba`: Make probability predictions (if supported by the model)
  - Request body: `[{{"feature1": value1, "feature2": value2, ...}}, ...]`
  - Response: `{{"probabilities": [[prob1_class1, prob1_class2, ...], [prob2_class1, prob2_class2, ...], ...]}}`

- `GET /health`: Health check
  - Response: `{{"status": "healthy"}}`

- `GET /metadata`: Get model metadata
  - Response: Model metadata

## Docker

You can also run the API using Docker:

```
docker build -t {api_name.lower().replace(" ", "-")} .
docker run -p {port}:{port} {api_name.lower().replace(" ", "-")}
```
"""
    
    def _generate_flask_dockerfile(self) -> str:
        """
        Generate Flask Dockerfile content.
        
        Returns:
            Content of Dockerfile.
        """
        return """
FROM python:3.9-slim

WORKDIR /app

COPY requirements.txt .
RUN pip install --no-cache-dir -r requirements.txt

COPY . .

CMD ["python", "app.py"]
"""
    
    def _generate_streamlit_app(self, model_path: str, app_name: str,
                               include_preprocessing: bool, preprocessor_path: Optional[str],
                               include_visualization: bool) -> str:
        """
        Generate Streamlit app.py content.
        
        Args:
            model_path: Path to the model file.
            app_name: Name of the application.
            include_preprocessing: Whether to include preprocessing.
            preprocessor_path: Path to the preprocessor file.
            include_visualization: Whether to include visualization.
            
        Returns:
            Content of app.py.
        """
        model_filename = os.path.basename(model_path)
        preprocessor_filename = os.path.basename(preprocessor_path) if preprocessor_path else None
        
        content = f"""
import os
import pickle
import json

import numpy as np
import pandas as pd
import streamlit as st
"""
        
        if include_visualization:
            content += """
import matplotlib.pyplot as plt
import seaborn as sns
"""
        
        content += f"""
# Set page title
st.set_page_config(page_title="{app_name}", layout="wide")

# App title
st.title("{app_name}")

# Load model
@st.cache_resource
def load_model():
    try:
        with open("{model_filename}", "rb") as f:
            loaded_obj = pickle.load(f)
        
        # Check if the loaded object is a dictionary with model and metadata
        if isinstance(loaded_obj, dict) and 'model' in loaded_obj:
            model = loaded_obj['model']
            metadata = loaded_obj.get('metadata', {{}})
        else:
            # Assume the loaded object is the model itself
            model = loaded_obj
            metadata = {{}}
        
        return model, metadata
    except Exception as e:
        st.error(f"Error loading model: {{e}}")
        return None, {{}}

model, metadata = load_model()
"""
        
        if include_preprocessing and preprocessor_path:
            content += f"""
# Load preprocessor
@st.cache_resource
def load_preprocessor():
    try:
        with open("{preprocessor_filename}", "rb") as f:
            loaded_obj = pickle.load(f)
        
        # Check if the loaded object is a dictionary with preprocessor and metadata
        if isinstance(loaded_obj, dict) and 'preprocessor' in loaded_obj:
            preprocessor = loaded_obj['preprocessor']
        else:
            # Assume the loaded object is the preprocessor itself
            preprocessor = loaded_obj
        
        return preprocessor
    except Exception as e:
        st.error(f"Error loading preprocessor: {{e}}")
        return None

preprocessor = load_preprocessor()
"""
        else:
            content += """
# No preprocessor
preprocessor = None
"""
        
        content += """
# Sidebar
st.sidebar.header("Model Information")

# Display model type
model_type = type(model).__name__
st.sidebar.write(f"**Model Type:** {model_type}")

# Display model parameters
st.sidebar.write("**Model Parameters:**")
if hasattr(model, 'get_params'):
    params = model.get_params()
    for param, value in params.items():
        st.sidebar.write(f"- {param}: {value}")

# Display metadata
if metadata:
    st.sidebar.write("**Model Metadata:**")
    for key, value in metadata.items():
        if not isinstance(value, (dict, list, np.ndarray, pd.DataFrame, pd.Series)):
            st.sidebar.write(f"- {key}: {value}")

# Main content
st.header("Make Predictions")

# Input method selection
input_method = st.radio("Select input method:", ["Manual Input", "Upload CSV"])

if input_method == "Manual Input":
    # Get feature names
    if hasattr(model, 'feature_names_in_'):
        feature_names = model.feature_names_in_
    elif 'feature_names' in metadata:
        feature_names = metadata['feature_names']
    else:
        # Ask user for feature names
        feature_names_input = st.text_input("Enter feature names (comma-separated):")
        feature_names = [name.strip() for name in feature_names_input.split(',')] if feature_names_input else []
    
    # Create input fields for each feature
    input_data = {}
    
    if feature_names:
        st.subheader("Enter feature values:")
        cols = st.columns(2)
        for i, feature in enumerate(feature_names):
            input_data[feature] = cols[i % 2].number_input(f"{feature}", value=0.0, key=f"feature_{i}")
    else:
        st.warning("No feature names available. Please upload a CSV file instead.")
    
    # Make prediction button
    if feature_names and st.button("Predict"):
        # Create DataFrame from input
        input_df = pd.DataFrame([input_data])
        
        # Apply preprocessor if available
        if preprocessor is not None:
            try:
                input_df = preprocessor.transform(input_df)
            except Exception as e:
                st.error(f"Error applying preprocessor: {e}")
        
        # Make prediction
        try:
            prediction = model.predict(input_df)
            st.success(f"Prediction: {prediction[0]}")
            
            # Show probability if available
            if hasattr(model, 'predict_proba'):
                probabilities = model.predict_proba(input_df)
                st.write("Prediction probabilities:")
                
                # Get class names if available
                if hasattr(model, 'classes_'):
                    class_names = model.classes_
                    prob_df = pd.DataFrame([probabilities[0]], columns=class_names)
                else:
                    prob_df = pd.DataFrame([probabilities[0]], columns=[f"Class {i}" for i in range(probabilities.shape[1])])
                
                st.dataframe(prob_df)
"""
        
        if include_visualization:
            content += """
                # Visualize probabilities
                if hasattr(model, 'predict_proba'):
                    fig, ax = plt.subplots(figsize=(10, 4))
                    
                    # Get class names if available
                    if hasattr(model, 'classes_'):
                        class_names = model.classes_
                    else:
                        class_names = [f"Class {i}" for i in range(probabilities.shape[1])]
                    
                    # Plot probabilities
                    sns.barplot(x=class_names, y=probabilities[0], ax=ax)
                    ax.set_title("Prediction Probabilities")
                    ax.set_ylabel("Probability")
                    ax.set_xlabel("Class")
                    
                    # Rotate x-axis labels if there are many classes
                    if len(class_names) > 3:
                        plt.xticks(rotation=45, ha='right')
                    
                    st.pyplot(fig)
"""
        
        content += """
        except Exception as e:
            st.error(f"Error making prediction: {e}")

elif input_method == "Upload CSV":
    # File uploader
    uploaded_file = st.file_uploader("Upload a CSV file", type=["csv"])
    
    if uploaded_file is not None:
        # Read CSV
        try:
            input_df = pd.read_csv(uploaded_file)
            st.write("Preview of uploaded data:")
            st.dataframe(input_df.head())
            
            # Make prediction button
            if st.button("Predict"):
                # Apply preprocessor if available
                if preprocessor is not None:
                    try:
                        processed_df = preprocessor.transform(input_df)
                    except Exception as e:
                        st.error(f"Error applying preprocessor: {e}")
                        processed_df = input_df
                else:
                    processed_df = input_df
                
                # Make predictions
                try:
                    predictions = model.predict(processed_df)
                    
                    # Create results DataFrame
                    results_df = input_df.copy()
                    results_df['prediction'] = predictions
                    
                    # Add probabilities if available
                    if hasattr(model, 'predict_proba'):
                        probabilities = model.predict_proba(processed_df)
                        
                        # Get class names if available
                        if hasattr(model, 'classes_'):
                            class_names = model.classes_
                            for i, class_name in enumerate(class_names):
                                results_df[f"probability_{class_name}"] = probabilities[:, i]
                        else:
                            for i in range(probabilities.shape[1]):
                                results_df[f"probability_class_{i}"] = probabilities[:, i]
                    
                    st.success("Predictions completed!")
                    st.write("Results:")
                    st.dataframe(results_df)
                    
                    # Download button
                    csv = results_df.to_csv(index=False)
                    st.download_button(
                        label="Download results as CSV",
                        data=csv,
                        file_name="predictions.csv",
                        mime="text/csv"
                    )
"""
        
        if include_visualization:
            content += """
                    # Visualize predictions
                    st.subheader("Visualization")
                    
                    # Check if it's a classification problem
                    if hasattr(model, 'classes_'):
                        # Plot prediction distribution
                        fig, ax = plt.subplots(figsize=(10, 6))
                        prediction_counts = results_df['prediction'].value_counts().sort_index()
                        sns.barplot(x=prediction_counts.index, y=prediction_counts.values, ax=ax)
                        ax.set_title("Prediction Distribution")
                        ax.set_ylabel("Count")
                        ax.set_xlabel("Predicted Class")
                        
                        # Rotate x-axis labels if there are many classes
                        if len(prediction_counts) > 3:
                            plt.xticks(rotation=45, ha='right')
                        
                        st.pyplot(fig)
                    else:
                        # Regression problem
                        fig, ax = plt.subplots(figsize=(10, 6))
                        sns.histplot(predictions, kde=True, ax=ax)
                        ax.set_title("Prediction Distribution")
                        ax.set_xlabel("Predicted Value")
                        ax.set_ylabel("Count")
                        st.pyplot(fig)
"""
        
        content += """
                except Exception as e:
                    st.error(f"Error making predictions: {e}")
        except Exception as e:
            st.error(f"Error reading CSV file: {e}")

# About section
st.sidebar.header("About")
st.sidebar.info(
    f"This is a Streamlit app for demonstrating the {app_name} model. "
    "It allows you to make predictions using the model either by manual input or by uploading a CSV file."
)
"""
        
        return content
    
    def _generate_streamlit_requirements(self) -> str:
        """
        Generate Streamlit requirements.txt content.
        
        Returns:
            Content of requirements.txt.
        """
        return """
streamlit>=1.10.0
pandas>=1.3.0
numpy>=1.20.0
scikit-learn>=0.24.0
matplotlib>=3.4.0
seaborn>=0.11.0
"""
    
    def _generate_streamlit_readme(self, app_name: str, port: int) -> str:
        """
        Generate Streamlit README.md content.
        
        Args:
            app_name: Name of the application.
            port: Port to bind the server.
            
        Returns:
            Content of README.md.
        """
        return f"""
# {app_name}

This is a Streamlit application for demonstrating a machine learning model.

## Installation

1. Clone this repository
2. Install dependencies:
   ```
   pip install -r requirements.txt
   ```

## Usage

### Running the Application

```
streamlit run app.py
```

The application will be available at http://localhost:{port}

### Features

- Make predictions using manual input or by uploading a CSV file
- View model information and parameters
- Visualize prediction results
- Download prediction results as CSV

## Docker

You can also run the application using Docker:

```
docker build -t {app_name.lower().replace(" ", "-")} .
docker run -p {port}:{port} {app_name.lower().replace(" ", "-")}
```
"""
    
    def _generate_streamlit_dockerfile(self) -> str:
        """
        Generate Streamlit Dockerfile content.
        
        Returns:
            Content of Dockerfile.
        """
        return """
FROM python:3.9-slim

WORKDIR /app

COPY requirements.txt .
RUN pip install --no-cache-dir -r requirements.txt

COPY . .

EXPOSE 8501

CMD ["streamlit", "run", "app.py", "--server.address=0.0.0.0"]
"""
    
    def _generate_dockerignore(self) -> str:
        """
        Generate .dockerignore content.
        
        Returns:
            Content of .dockerignore.
        """
        return """
__pycache__/
*.py[cod]
*$py.class
.env
.venv
env/
venv/
ENV/
.git
.gitignore
.idea
.vscode
"""



"""
Environment setup for SBYB Scaffolding.

This module provides functionality to set up development environments
for ML projects, including virtual environments, package installation,
and environment configuration.
"""

from typing import Any, Dict, List, Optional, Union
import os
import sys
import subprocess
import platform
import shutil
from pathlib import Path

from sbyb.core.base import SBYBComponent
from sbyb.core.config import Config
from sbyb.core.exceptions import ScaffoldingError


class EnvironmentSetup(SBYBComponent):
    """
    Environment setup for ML projects.
    
    This component provides functionality to set up development environments
    for ML projects, including virtual environments, package installation,
    and environment configuration.
    """
    
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """
        Initialize the environment setup.
        
        Args:
            config: Configuration dictionary for the environment setup.
        """
        super().__init__(config)
    
    def create_virtual_env(self, output_dir: str, env_name: str = "venv",
                          python_version: Optional[str] = None) -> str:
        """
        Create a Python virtual environment.
        
        Args:
            output_dir: Directory to create the virtual environment in.
            env_name: Name of the virtual environment.
            python_version: Python version to use.
            
        Returns:
            Path to the virtual environment.
        """
        env_path = os.path.join(output_dir, env_name)
        
        # Check if virtual environment already exists
        if os.path.exists(env_path):
            raise ScaffoldingError(f"Virtual environment already exists at {env_path}")
        
        # Determine Python executable
        if python_version:
            # Try to find specific Python version
            python_cmd = f"python{python_version}"
            try:
                subprocess.run([python_cmd, "--version"], check=True, capture_output=True)
            except (subprocess.SubprocessError, FileNotFoundError):
                raise ScaffoldingError(f"Python version {python_version} not found")
        else:
            # Use current Python
            python_cmd = sys.executable
        
        # Create virtual environment
        try:
            subprocess.run([python_cmd, "-m", "venv", env_path], check=True)
        except subprocess.SubprocessError as e:
            raise ScaffoldingError(f"Failed to create virtual environment: {str(e)}")
        
        return env_path
    
    def install_requirements(self, env_path: str, requirements_path: str) -> None:
        """
        Install requirements in a virtual environment.
        
        Args:
            env_path: Path to the virtual environment.
            requirements_path: Path to the requirements.txt file.
        """
        # Determine pip executable
        if platform.system() == "Windows":
            pip_path = os.path.join(env_path, "Scripts", "pip")
        else:
            pip_path = os.path.join(env_path, "bin", "pip")
        
        # Check if pip exists
        if not os.path.exists(pip_path):
            raise ScaffoldingError(f"Pip not found in virtual environment at {pip_path}")
        
        # Check if requirements file exists
        if not os.path.exists(requirements_path):
            raise ScaffoldingError(f"Requirements file not found at {requirements_path}")
        
        # Install requirements
        try:
            subprocess.run([pip_path, "install", "-r", requirements_path], check=True)
        except subprocess.SubprocessError as e:
            raise ScaffoldingError(f"Failed to install requirements: {str(e)}")
    
    def install_package(self, env_path: str, package_name: str,
                       version: Optional[str] = None) -> None:
        """
        Install a package in a virtual environment.
        
        Args:
            env_path: Path to the virtual environment.
            package_name: Name of the package to install.
            version: Version of the package to install.
        """
        # Determine pip executable
        if platform.system() == "Windows":
            pip_path = os.path.join(env_path, "Scripts", "pip")
        else:
            pip_path = os.path.join(env_path, "bin", "pip")
        
        # Check if pip exists
        if not os.path.exists(pip_path):
            raise ScaffoldingError(f"Pip not found in virtual environment at {pip_path}")
        
        # Prepare package specification
        if version:
            package_spec = f"{package_name}=={version}"
        else:
            package_spec = package_name
        
        # Install package
        try:
            subprocess.run([pip_path, "install", package_spec], check=True)
        except subprocess.SubprocessError as e:
            raise ScaffoldingError(f"Failed to install package {package_spec}: {str(e)}")
    
    def install_development_mode(self, env_path: str, package_dir: str) -> None:
        """
        Install a package in development mode.
        
        Args:
            env_path: Path to the virtual environment.
            package_dir: Path to the package directory.
        """
        # Determine pip executable
        if platform.system() == "Windows":
            pip_path = os.path.join(env_path, "Scripts", "pip")
        else:
            pip_path = os.path.join(env_path, "bin", "pip")
        
        # Check if pip exists
        if not os.path.exists(pip_path):
            raise ScaffoldingError(f"Pip not found in virtual environment at {pip_path}")
        
        # Check if package directory exists
        if not os.path.exists(package_dir):
            raise ScaffoldingError(f"Package directory not found at {package_dir}")
        
        # Install package in development mode
        try:
            subprocess.run([pip_path, "install", "-e", package_dir], check=True)
        except subprocess.SubprocessError as e:
            raise ScaffoldingError(f"Failed to install package in development mode: {str(e)}")
    
    def create_conda_env(self, env_name: str, environment_file: str) -> str:
        """
        Create a conda environment.
        
        Args:
            env_name: Name of the conda environment.
            environment_file: Path to the environment.yml file.
            
        Returns:
            Name of the created conda environment.
        """
        # Check if conda is available
        try:
            subprocess.run(["conda", "--version"], check=True, capture_output=True)
        except (subprocess.SubprocessError, FileNotFoundError):
            raise ScaffoldingError("Conda not found. Please install Anaconda or Miniconda.")
        
        # Check if environment file exists
        if not os.path.exists(environment_file):
            raise ScaffoldingError(f"Environment file not found at {environment_file}")
        
        # Create conda environment
        try:
            subprocess.run(["conda", "env", "create", "-n", env_name, "-f", environment_file], check=True)
        except subprocess.SubprocessError as e:
            raise ScaffoldingError(f"Failed to create conda environment: {str(e)}")
        
        return env_name
    
    def setup_jupyter_kernel(self, env_path: str, kernel_name: str,
                            display_name: Optional[str] = None) -> None:
        """
        Set up a Jupyter kernel for a virtual environment.
        
        Args:
            env_path: Path to the virtual environment.
            kernel_name: Name of the kernel.
            display_name: Display name of the kernel.
        """
        # Determine Python executable
        if platform.system() == "Windows":
            python_path = os.path.join(env_path, "Scripts", "python")
        else:
            python_path = os.path.join(env_path, "bin", "python")
        
        # Check if Python exists
        if not os.path.exists(python_path):
            raise ScaffoldingError(f"Python not found in virtual environment at {python_path}")
        
        # Set display name
        if display_name is None:
            display_name = kernel_name
        
        # Install ipykernel if not already installed
        try:
            subprocess.run([python_path, "-m", "pip", "install", "ipykernel"], check=True)
        except subprocess.SubprocessError as e:
            raise ScaffoldingError(f"Failed to install ipykernel: {str(e)}")
        
        # Set up Jupyter kernel
        try:
            subprocess.run([
                python_path, "-m", "ipykernel", "install",
                "--user", "--name", kernel_name, "--display-name", display_name
            ], check=True)
        except subprocess.SubprocessError as e:
            raise ScaffoldingError(f"Failed to set up Jupyter kernel: {str(e)}")
    
    def setup_git_repo(self, project_dir: str, init_commit: bool = True) -> None:
        """
        Set up a Git repository for a project.
        
        Args:
            project_dir: Path to the project directory.
            init_commit: Whether to create an initial commit.
        """
        # Check if git is available
        try:
            subprocess.run(["git", "--version"], check=True, capture_output=True)
        except (subprocess.SubprocessError, FileNotFoundError):
            raise ScaffoldingError("Git not found. Please install Git.")
        
        # Check if project directory exists
        if not os.path.exists(project_dir):
            raise ScaffoldingError(f"Project directory not found at {project_dir}")
        
        # Initialize git repository
        try:
            subprocess.run(["git", "init"], check=True, cwd=project_dir)
        except subprocess.SubprocessError as e:
            raise ScaffoldingError(f"Failed to initialize git repository: {str(e)}")
        
        # Create initial commit if requested
        if init_commit:
            try:
                subprocess.run(["git", "add", "."], check=True, cwd=project_dir)
                subprocess.run(["git", "commit", "-m", "Initial commit"], check=True, cwd=project_dir)
            except subprocess.SubprocessError as e:
                raise ScaffoldingError(f"Failed to create initial commit: {str(e)}")
    
    def setup_pre_commit_hooks(self, project_dir: str) -> None:
        """
        Set up pre-commit hooks for a project.
        
        Args:
            project_dir: Path to the project directory.
        """
        # Check if pre-commit is available
        try:
            subprocess.run(["pre-commit", "--version"], check=True, capture_output=True)
        except (subprocess.SubprocessError, FileNotFoundError):
            raise ScaffoldingError("pre-commit not found. Please install pre-commit.")
        
        # Check if project directory exists
        if not os.path.exists(project_dir):
            raise ScaffoldingError(f"Project directory not found at {project_dir}")
        
        # Create pre-commit config file
        pre_commit_config = """repos:
-   repo: https://github.com/pre-commit/pre-commit-hooks
    rev: v4.4.0
    hooks:
    -   id: trailing-whitespace
    -   id: end-of-file-fixer
    -   id: check-yaml
    -   id: check-added-large-files

-   repo: https://github.com/psf/black
    rev: 23.3.0
    hooks:
    -   id: black

-   repo: https://github.com/pycqa/isort
    rev: 5.12.0
    hooks:
    -   id: isort

-   repo: https://github.com/pycqa/flake8
    rev: 6.0.0
    hooks:
    -   id: flake8
        additional_dependencies: [flake8-docstrings]
"""
        
        # Write pre-commit config file
        pre_commit_path = os.path.join(project_dir, ".pre-commit-config.yaml")
        with open(pre_commit_path, "w") as f:
            f.write(pre_commit_config)
        
        # Install pre-commit hooks
        try:
            subprocess.run(["pre-commit", "install"], check=True, cwd=project_dir)
        except subprocess.SubprocessError as e:
            raise ScaffoldingError(f"Failed to install pre-commit hooks: {str(e)}")
    
    def setup_vscode_settings(self, project_dir: str) -> None:
        """
        Set up VS Code settings for a project.
        
        Args:
            project_dir: Path to the project directory.
        """
        # Check if project directory exists
        if not os.path.exists(project_dir):
            raise ScaffoldingError(f"Project directory not found at {project_dir}")
        
        # Create .vscode directory
        vscode_dir = os.path.join(project_dir, ".vscode")
        os.makedirs(vscode_dir, exist_ok=True)
        
        # Create settings.json
        settings = {
            "python.linting.enabled": True,
            "python.linting.flake8Enabled": True,
            "python.linting.pylintEnabled": False,
            "python.formatting.provider": "black",
            "editor.formatOnSave": True,
            "editor.codeActionsOnSave": {
                "source.organizeImports": True
            },
            "python.testing.pytestEnabled": True,
            "python.testing.unittestEnabled": False,
            "python.testing.nosetestsEnabled": False,
            "python.testing.pytestArgs": [
                "tests"
            ]
        }
        
        # Write settings.json
        settings_path = os.path.join(vscode_dir, "settings.json")
        with open(settings_path, "w") as f:
            import json
            json.dump(settings, f, indent=4)
        
        # Create launch.json
        launch = {
            "version": "0.2.0",
            "configurations": [
                {
                    "name": "Python: Current File",
                    "type": "python",
                    "request": "launch",
                    "program": "${file}",
                    "console": "integratedTerminal"
                }
            ]
        }
        
        # Write launch.json
        launch_path = os.path.join(vscode_dir, "launch.json")
        with open(launch_path, "w") as f:
            import json
            json.dump(launch, f, indent=4)
    
    def setup_project_environment(self, project_dir: str, env_type: str = "venv",
                                 python_version: Optional[str] = None,
                                 setup_git: bool = True,
                                 setup_jupyter: bool = True,
                                 setup_vscode: bool = True) -> Dict[str, Any]:
        """
        Set up a complete project environment.
        
        Args:
            project_dir: Path to the project directory.
            env_type: Type of environment to create ('venv' or 'conda').
            python_version: Python version to use.
            setup_git: Whether to set up Git.
            setup_jupyter: Whether to set up Jupyter.
            setup_vscode: Whether to set up VS Code.
            
        Returns:
            Dictionary with environment information.
        """
        # Check if project directory exists
        if not os.path.exists(project_dir):
            raise ScaffoldingError(f"Project directory not found at {project_dir}")
        
        # Get project name from directory
        project_name = os.path.basename(os.path.abspath(project_dir))
        
        # Create environment
        env_info = {}
        if env_type == "venv":
            # Create virtual environment
            env_path = self.create_virtual_env(project_dir, "venv", python_version)
            env_info["env_type"] = "venv"
            env_info["env_path"] = env_path
            
            # Install requirements if requirements.txt exists
            requirements_path = os.path.join(project_dir, "requirements.txt")
            if os.path.exists(requirements_path):
                self.install_requirements(env_path, requirements_path)
            
            # Install package in development mode if setup.py exists
            setup_path = os.path.join(project_dir, "setup.py")
            if os.path.exists(setup_path):
                self.install_development_mode(env_path, project_dir)
            
            # Set up Jupyter kernel if requested
            if setup_jupyter:
                self.setup_jupyter_kernel(env_path, project_name, f"{project_name} (venv)")
        
        elif env_type == "conda":
            # Check if environment.yml exists
            env_file = os.path.join(project_dir, "environment.yml")
            if not os.path.exists(env_file):
                raise ScaffoldingError(f"Environment file not found at {env_file}")
            
            # Create conda environment
            env_name = self.create_conda_env(project_name, env_file)
            env_info["env_type"] = "conda"
            env_info["env_name"] = env_name
            
            # Set up Jupyter kernel if requested (conda environments automatically register kernels)
        
        else:
            raise ScaffoldingError(f"Unsupported environment type: {env_type}")
        
        # Set up Git if requested
        if setup_git:
            self.setup_git_repo(project_dir)
            
            # Check if .gitignore exists, create if not
            gitignore_path = os.path.join(project_dir, ".gitignore")
            if not os.path.exists(gitignore_path):
                from sbyb.scaffolding.config import ConfigGenerator
                config_gen = ConfigGenerator()
                config_gen.generate_gitignore(gitignore_path)
        
        # Set up VS Code if requested
        if setup_vscode:
            self.setup_vscode_settings(project_dir)
        
        return env_info
